import copy
import hashlib
import json
from datetime import datetime, timezone
from types import SimpleNamespace

import httpx
import pytest
from polyfactory.factories.pydantic_factory import ModelFactory

from fbnconfig import identity, lumi, scheduler
from fbnconfig import workflows as wf
from fbnconfig.workflows import LuminesceView, SchedulerJob

TEST_BASE = "https://foo.lusid.com"

WORKER_CONFIGS = [
    (wf.Sleep(), {"type": "Sleep"}, "0"),
    (wf.Fail(), {"type": "Fail"}, "0"),
    (wf.HealthCheck(url="a.com"), {"type": "HealthCheck", "url": "a.com"}, "0"),
    (
        wf.LuminesceView(view=lumi.ViewRef(id="a", provider="Views.Unit.Something")),
        {"type": "LuminesceView", "name": "Views.Unit.Something"},
        "1",
    ),
    (
        wf.LuminesceView(
            view=lumi.ViewResource(
                id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
            )
        ),
        {"type": "LuminesceView", "name": "Views.Unit.Test"},
        "5c9205c16d8e2036aad15d4e827e7f0a6f2cbd159c2b2bf2cf7f7123a5ac0b65",
    ),
    (
        wf.SchedulerJob(job=scheduler.JobRef(id="A", scope="B", code="C")),
        {"type": "SchedulerJob", "jobId": {"scope": "B", "code": "C"}},
        "ABCD",
    ),
    (
        wf.SchedulerJob(
            job=scheduler.JobResource(
                id="A",
                scope="jobscope",
                code="jobcode",
                image=scheduler.ImageRef(id="img", dest_name="img-name", dest_tag="sometag"),
                name="A job",
                description="A",
            )
        ),
        {"type": "SchedulerJob", "jobId": {"scope": "jobscope", "code": "jobcode"}},
        "ce31cf3f48906e757c2daccfae062dc17d04bc7d66cfd04e8867c504eb4e8490",
    ),
]


class WorkerResourceFactory(ModelFactory[wf.WorkerResource]): ...


class TaskDefFactory(ModelFactory[wf.TaskDefinitionResource]): ...


class TaskDefRefFactory(ModelFactory[wf.TaskDefinitionRef]): ...


class UserFactory(ModelFactory[identity.UserResource]): ...


class EventHandlerFactory(ModelFactory[wf.EventHandlerResource]): ...


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWorkerResource:
    base_url = TEST_BASE
    workers_url = "/workflow/api/workers"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @pytest.mark.parametrize("worker_config, worker_config_json, version", WORKER_CONFIGS)
    def test_create_worker(self, respx_mock, worker_config, worker_config_json, version):
        # since tests modify the versions on the configs, copy in to allow parallel runs
        # deepcopy as changes are made to the underlying object rather than worker config itself
        config = copy.deepcopy(worker_config)
        # GIVEN Resource does not exist
        wr = wf.WorkerResource(
            id="wr1",
            scope="somescope",
            code="workercode",
            display_name="name",
            description="descriptions",
            worker_configuration=config,
        )
        if isinstance(config, wf.SchedulerJob) and isinstance(config.job, scheduler.JobRef):
            config.job._content_hash = "ABCD"
        if isinstance(config, LuminesceView) and isinstance(config.view, lumi.ViewRef):
            config.view._version = version
        respx_mock.post(self.workers_url).mock(
            side_effect=[
                httpx.Response(
                    status_code=201,
                    json={
                        "id": {"scope": "somescope", "code": "workercode"},
                        "displayName": "name",
                        "description": "descriptions",
                        "workerConfiguration": worker_config_json,
                        "version": {"asAtCreated": "somedate", "asAtVersionNumber": 1},
                        "parameters": [],
                        "resultFields": [],
                    },
                )
            ]
        )

        # WHEN Worker Resource is created
        new_state = wr.create(self.client)
        # THEN State returned as expected
        assert new_state == {
            "id": "wr1",
            "scope": "somescope",
            "code": "workercode",
            "worker_version": version,
        }

        request: httpx.Request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == self.workers_url
        assert f"https://{request.url.host}" == self.base_url
        assert json.loads(request.content) == {
            "id": {"scope": "somescope", "code": "workercode"},
            "displayName": "name",
            "description": "descriptions",
            "workerConfiguration": worker_config_json,
        }

    @pytest.mark.parametrize("worker_config, worker_config_json, version", WORKER_CONFIGS)
    def test_update_no_change(self, respx_mock, worker_config, worker_config_json, version):
        config = copy.deepcopy(worker_config)
        # GIVEN Resource exists
        respx_mock.get(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "test", "code": "testWorker"},
                        "displayName": "DoWork",
                        "description": "Worker that does some work",
                        "workerConfiguration": worker_config_json,
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 2,
                        },
                        "parameters": [],
                        "resultFields": [
                            {
                                "name": "AccountingMethod",
                                "type": "String",
                                "displayName": "Accounting Method",
                            }
                        ],
                        "links": [{"method": "GET"}],
                    },
                )
            ]
        )

        old_state = SimpleNamespace(id="wr1", scope="test", code="testWorker", worker_version=version)
        # AND No Change
        wr = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoWork",
            description="Worker that does some work",
            worker_configuration=config,
        )

        if isinstance(config, wf.LuminesceView) and isinstance(config.view, lumi.ViewRef):
            config.view._version = version

        if isinstance(config, wf.SchedulerJob) and isinstance(config.job, scheduler.JobRef):
            config.job._content_hash = version

        new_state = wr.update(self.client, old_state)
        # Then no actions
        assert new_state is None

    @pytest.mark.parametrize("worker_config, worker_config_json, version", WORKER_CONFIGS)
    def test_update_identifier(self, respx_mock, worker_config, worker_config_json, version):
        # GIVEN Worker exists
        # WHen updating scope or code
        # THEN Worker is deleted and recreated
        config = copy.deepcopy(worker_config)
        respx_mock.delete(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[httpx.Response(200, json={"asAt": "asAt deleted"})]
        )

        respx_mock.post(self.workers_url).mock(
            side_effect=[
                httpx.Response(
                    status_code=201,
                    json={
                        "id": {"scope": "test2", "code": "testWorker2"},
                        "displayName": "DoWork",
                        "workerConfiguration": worker_config_json,
                    },
                )
            ]
        )
        old_state = SimpleNamespace(id="wr1", scope="test", code="testWorker", worker_version=version)
        wr = wf.WorkerResource(
            id="wr1",
            scope="test2",
            code="testWorker2",
            display_name="DoWork",
            worker_configuration=config,
        )

        if isinstance(config, wf.LuminesceView) and isinstance(config.view, lumi.ViewRef):
            config.view._version = version
        if isinstance(config, wf.SchedulerJob) and isinstance(config.job, scheduler.JobRef):
            config.job._content_hash = version
        new_state = wr.update(self.client, old_state)
        assert new_state == {
            "id": "wr1",
            "scope": "test2",
            "code": "testWorker2",
            "worker_version": version,
        }

        # and the post is made to create the job
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert json.loads(request.content) == {
            "id": {"scope": "test2", "code": "testWorker2"},
            "displayName": "DoWork",
            "workerConfiguration": worker_config_json,
        }

    def test_update_fields_locally(self, respx_mock):
        # GIVEN Worker exists, but it has not changed from local
        respx_mock.get(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "test", "code": "testWorker"},
                        "displayName": "DoWork",
                        "description": "some description",
                        "workerConfiguration": {
                            "type": "SchedulerJob",
                            "jobId": {"scope": "jobscope", "code": "jobcode"},
                        },
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 3,
                        },
                        "parameters": {},
                        "resultFields": {},
                        "links": {},
                    },
                )
            ]
        )

        respx_mock.put(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    status_code=200,
                    json={
                        "displayName": "DoSomeDifferentWork",
                        "description": "New description",
                        "workerConfiguration": {
                            "type": "SchedulerJob",
                            "jobId": {"scope": "jobscope", "code": "jobcode"},
                        },
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 4,
                        },
                    },
                )
            ]
        )

        old_state = SimpleNamespace(
            id="wr1", scope="test", code="testWorker", worker_version="someworkerhash"
        )
        ref = scheduler.JobRef(id="A", scope="jobscope", code="jobcode")
        # WHEN The worker name and descriptions change
        # But not the worker config
        wr = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.SchedulerJob(job=ref),
        )

        ref._content_hash = "ABCD"

        new_state = wr.update(self.client, old_state)
        # THEN local hash and remote version update, but worker config version stays the same
        assert new_state == {
            "id": "wr1",
            "scope": "test",
            "code": "testWorker",
            "worker_version": "ABCD",
        }

    def test_remote_updates_fields(self, respx_mock):
        # GIVEN Worker exists, but it has changed from local
        respx_mock.get(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "test", "code": "testWorker"},
                        "displayName": "DoSomeOtherWork",
                        "description": "SomeOtherDescription",
                        "workerConfiguration": {
                            "type": "SchedulerJob",
                            "jobId": {"scope": "jobscope", "code": "jobcode"},
                        },
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 4,
                        },
                        "parameters": {},
                        "resultFields": {},
                        "links": {},
                    },
                )
            ]
        )
        respx_mock.put(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    status_code=200,
                    json={
                        "displayName": "Some name",
                        "description": "Some description",
                        "workerConfiguration": {
                            "type": "SchedulerJob",
                            "jobId": {"scope": "jobscope", "code": "jobcode"},
                        },
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 5,
                        },
                    },
                )
            ]
        )
        ref = lumi.ViewRef(id="A", provider="views.test.view1")
        ref._version = "1"
        old_state = SimpleNamespace(
            id="wr1", scope="test", code="testWorker", worker_version="someworkerhash"
        )

        # WHEN local has not changed but remote version has
        wr = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="Some name",
            description="Some description",
            worker_configuration=wf.LuminesceView(view=ref),
        )

        ref._version = "ABCD"
        new_state = wr.update(self.client, old_state)
        # THEN local hash and worker config version stays the same but remote version updates
        assert new_state == {
            "id": "wr1",
            "scope": "test",
            "code": "testWorker",
            "worker_version": "ABCD",
        }

    @pytest.mark.parametrize("worker_config, worker_config_json, version", WORKER_CONFIGS)
    def test_update_lumi_scheduler_worker_config(
        self, respx_mock, worker_config, worker_config_json, version
    ):
        config = copy.deepcopy(worker_config)
        respx_mock.get(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "test", "code": "testWorker"},
                        "displayName": "DoSomeOtherWork",
                        "description": "SomeOtherDescription",
                        "workerConfiguration": worker_config_json,
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 1,
                        },
                        "parameters": {},
                        "resultFields": {},
                        "links": {},
                    },
                )
            ]
        )

        respx_mock.put(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[
                httpx.Response(
                    status_code=200,
                    json={
                        "displayName": "Some name",
                        "description": "Some description",
                        "workerConfiguration": worker_config_json,
                        "version": {
                            "asAtCreated": "2023-10-09T12:48:27.6157190+00:00",
                            "asAtVersionNumber": 5,
                        },
                    },
                )
            ]
        )
        old_state = SimpleNamespace(id="wr1", scope="test", code="testWorker", worker_version="abc")

        wr = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="Some name",
            description="Some description",
            worker_configuration=config,
        )

        if isinstance(config, LuminesceView) and isinstance(config.view, lumi.ViewRef):
            config.view._version = version
        if isinstance(config, SchedulerJob) and isinstance(config.job, scheduler.JobRef):
            config.job._content_hash = version

        new_state = wr.update(self.client, old_state)
        # THEN local hash and worker config version stays the same but remote version updates
        assert new_state == {
            "id": "wr1",
            "scope": "test",
            "code": "testWorker",
            "worker_version": version,
        }

    def test_delete(self, respx_mock):
        respx_mock.delete(f"{self.workers_url}/test/testWorker").mock(
            side_effect=[httpx.Response(200, json={"asAt": "asAt deleted"})]
        )

        old_state = SimpleNamespace(id="wr1", scope="test", code="testWorker")
        wr = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.LuminesceView(
                view=lumi.ViewResource(
                    id="my-view",
                    provider="Views.Unit.Test",
                    sql="select * from foo;",
                    description="a view",
                )
            ),
        )
        assert wr.delete(self.client, old_state) is None

        request = respx_mock.calls.last.request
        assert request.method == "DELETE"

    @staticmethod
    def test_deps():
        job_ref = scheduler.JobRef(id="A", scope="B", code="C")
        job = scheduler.JobResource(
            id="A",
            scope="jobscope",
            code="jobcode",
            image=scheduler.ImageRef(id="img", dest_name="img-name", dest_tag="sometag"),
            name="A job",
            description="A",
        )

        view_ref = lumi.ViewRef(id="abc", provider="Views.Unit.Ref")

        view_resource = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )

        wr_lumi_view = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.LuminesceView(view=view_resource),
        )

        deps = wr_lumi_view.deps()
        assert deps == [view_resource]

        wr_lumi_ref = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.LuminesceView(view=view_ref),
        )
        deps = wr_lumi_ref.deps()
        assert deps == [view_ref]

        wr_healthcheck = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.HealthCheck(url="a.com"),
        )
        deps = wr_healthcheck.deps()
        assert deps == []

        wr_fail = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.Fail(),
        )
        deps = wr_fail.deps()
        assert deps == []

        wr_sleep = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.Sleep(),
        )
        deps = wr_sleep.deps()
        assert deps == []

        wr_job = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.SchedulerJob(job=job),
        )
        deps = wr_job.deps()
        assert deps == [job]

        wr_job_ref = wf.WorkerResource(
            id="wr1",
            scope="test",
            code="testWorker",
            display_name="DoSomeDifferentWork",
            description="New description",
            worker_configuration=wf.SchedulerJob(job=job_ref),
        )
        deps = wr_job_ref.deps()
        assert deps == [job_ref]


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWorkerRef:
    base_url = TEST_BASE
    workers_url = "/workflow/api/workers"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get(f"{self.workers_url}/w1/w2").mock(
            side_effect=[httpx.Response(200, json={"id": {"scope": "w1", "code": "w2"}})]
        )

        sut = wf.WorkerRef(id="one", scope="w1", code="w2")
        # when we call attach
        sut.attach(self.client)
        # then a get request was made with the scope passed as a parameter
        req = respx_mock.calls.last.request
        assert req.url == f"{self.base_url + self.workers_url}/w1/w2"

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get(f"{self.workers_url}/w1/w2").mock(
            side_effect=[httpx.Response(404, json={"name": "WorkerNotFound"})]
        )
        sut = wf.WorkerRef(id="two", scope="w1", code="w2")

        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)

        assert "Worker w1/w2 does not exist" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get(f"{self.workers_url}/w1/w2").mock(side_effect=[httpx.Response(400, json={})])
        client = self.client
        sut = wf.WorkerRef(id="three", scope="w1", code="w2")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeTaskDefinitionRef:
    base_url = TEST_BASE
    task_defs_url = "/workflow/api/taskdefinitions"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get(f"{self.task_defs_url}/t1/t2").mock(
            side_effect=[httpx.Response(200, json={"id": {"scope": "t1", "code": "t2"}})]
        )

        sut = wf.TaskDefinitionRef(id="one", scope="t1", code="t2")
        # when we call attach
        sut.attach(self.client)
        req = respx_mock.calls.last.request
        assert req.url == f"{self.base_url + self.task_defs_url}/t1/t2"

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get(f"{self.task_defs_url}/t1/t2").mock(
            side_effect=[httpx.Response(404, json={"name": "TaskDefinitionNotFound"})]
        )
        sut = wf.TaskDefinitionRef(id="two", scope="t1", code="t2")

        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)

        assert "Task definition t1/t2 does not exist" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get(f"{self.task_defs_url}/t1/t2").mock(side_effect=[httpx.Response(400, json={})])
        client = self.client
        sut = wf.TaskDefinitionRef(id="three", scope="t1", code="t2")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


class DescribeTaskDefinitionSerialization:
    @staticmethod
    def test_initial_state():
        ex = wf.InitialState(name="test")
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "test",
            "requiredFields": [],
        }
        ex = wf.InitialState(
            name=wf.TaskStateDefinition(name="test2"),
            required_fields=["a", wf.TaskFieldDefinition(name="b", type="String")],
        )
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "test2",
            "requiredFields": ["a", "b"],
        }

        ex = wf.InitialState(name="test", required_fields=None)
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {"name": "test"}

        ex = wf.InitialState(name="test", required_fields=[])
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True, exclude_defaults=True) == {
            "name": "test",
            "requiredFields": [],
        }

    @staticmethod
    def test_trigger_definition():
        ex = wf.TriggerDefinition(name="sometrigger", type="External")
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "sometrigger",
            "trigger": {"type": "External"},
        }

    @staticmethod
    def test_field_mapping():
        ex = wf.FieldMapping(map_from="something")
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {"mapFrom": "something"}

        ex = wf.FieldMapping(set_to="somethingelse")
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {"setTo": "somethingelse"}

    @staticmethod
    def test_create_child_tasks_action():
        user_ref = identity.UserRef(id="s", login="a")
        user_ref.user_id = "myrefuser"
        ex = wf.ActionDefinition(
            name="someaction",
            action_details=wf.CreateChildTasksAction(
                child_task_configurations=[
                    wf.ChildTaskConfiguration(
                        task_definition=wf.TaskDefinitionRef(id="1", scope="somescope", code="somecode")
                    )
                ]
            ),
            run_as_user_id=user_ref,
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "someaction",
            "actionDetails": {
                "type": "CreateChildTasks",
                "childTaskConfigurations": [
                    {"taskDefinitionId": {"scope": "somescope", "code": "somecode"}}
                ],
            },
            "runAsUserId": "myrefuser",
        }

        ex = wf.ActionDefinition(
            name="someaction",
            action_details=wf.CreateChildTasksAction(
                child_task_configurations=[
                    wf.ChildTaskConfiguration(
                        task_definition=wf.TaskDefinitionRef(id="1", scope="somescope", code="somecode")
                    )
                ]
            ),
            run_as_user_id="myuserid",
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "someaction",
            "actionDetails": {
                "type": "CreateChildTasks",
                "childTaskConfigurations": [
                    {"taskDefinitionId": {"scope": "somescope", "code": "somecode"}}
                ],
            },
            "runAsUserId": "myuserid",
        }
        usr = identity.UserResource(
            id="something",
            first_name="s",
            last_name="b",
            email_address="d",
            login="d",
            type=identity.UserType.SERVICE,
        )

        usr.user_id = "myserviceuserid"
        ex = wf.ActionDefinition(
            name="mysecondaction",
            action_details=wf.CreateChildTasksAction(
                child_task_configurations=[
                    wf.ChildTaskConfiguration(
                        task_definition=wf.TaskDefinitionRef(id="1", scope="somescope", code="somecode"),
                        task_definition_as_at="2018-01-01T00:00:00+00:00",
                        initial_trigger=wf.TriggerDefinition(name="sometrigger", type="External"),
                        child_task_fields={"somefield": wf.FieldMapping(map_from="somethingelse")},
                        map_stacking_key_from="somewhere_else",
                    ),
                    wf.ChildTaskConfiguration(
                        task_definition=wf.TaskDefinitionRef(
                            id="1", scope="somescope", code="somecode2"
                        ),
                        task_definition_as_at=datetime(year=2018, month=1, day=1, tzinfo=timezone.utc),
                        initial_trigger="sometrigger2",
                        child_task_fields={},
                    ),
                ]
            ),
            run_as_user_id=usr,
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "mysecondaction",
            "actionDetails": {
                "type": "CreateChildTasks",
                "childTaskConfigurations": [
                    {
                        "taskDefinitionId": {"scope": "somescope", "code": "somecode"},
                        "taskDefinitionAsAt": "2018-01-01T00:00:00+00:00",
                        "initialTrigger": "sometrigger",
                        "childTaskFields": {"somefield": {"mapFrom": "somethingelse"}},
                        "mapStackingKeyFrom": "somewhere_else",
                    },
                    {
                        "taskDefinitionId": {"scope": "somescope", "code": "somecode2"},
                        "taskDefinitionAsAt": "2018-01-01T00:00:00+00:00",
                        "initialTrigger": "sometrigger2",
                        "childTaskFields": {},
                    },
                ],
            },
            "runAsUserId": "myserviceuserid",
        }

    @staticmethod
    def test_worker_status_triggers():
        trigger = wf.TriggerDefinition(name="mytrigger", type="External")
        ex = wf.WorkerStatusTriggers(
            started="trigger1",
            completed_with_results="trigger2",
            completed_no_results="trigger3",
            failed_to_start="trigger4",
            failed_to_complete="trigger5",
        )
        assert ex.model_dump(mode="json", by_alias=True) == {
            "started": "trigger1",
            "completedWithResults": "trigger2",
            "completedNoResults": "trigger3",
            "failedToStart": "trigger4",
            "failedToComplete": "trigger5",
        }

        ex = wf.WorkerStatusTriggers(
            started=trigger,
            completed_with_results=trigger,
            completed_no_results=trigger,
            failed_to_start=trigger,
            failed_to_complete=trigger,
        )

        assert ex.model_dump(mode="json", by_alias=True) == {
            "started": "mytrigger",
            "completedWithResults": "mytrigger",
            "completedNoResults": "mytrigger",
            "failedToStart": "mytrigger",
            "failedToComplete": "mytrigger",
        }

        ex = wf.WorkerStatusTriggers(completed_with_results="justonetrigger")
        assert ex.model_dump(mode="json", by_alias=True) == {
            "completedWithResults": "justonetrigger",
            "started": None,
            "completedNoResults": None,
            "failedToStart": None,
            "failedToComplete": None,
        }

    @staticmethod
    def test_resultant_child_task_configuration():
        child_task_config = wf.ChildTaskConfiguration(
            task_definition=wf.TaskDefinitionRef(id="1", scope="somescope", code="somecode"),
            child_task_fields={"somefield": wf.FieldMapping(map_from="somethingelse")},
        )
        ex = wf.ResultantChildTaskConfiguration(
            child_task_configuration=child_task_config, result_matching_pattern="something"
        )
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "taskDefinitionId": {"scope": "somescope", "code": "somecode"},
            "childTaskFields": {"somefield": {"mapFrom": "somethingelse"}},
            "resultMatchingPattern": {"filter": "something"},
        }

        ex = wf.ResultantChildTaskConfiguration(child_task_configuration=child_task_config)
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "taskDefinitionId": {"scope": "somescope", "code": "somecode"},
            "childTaskFields": {"somefield": {"mapFrom": "somethingelse"}},
        }

        field = wf.TaskFieldDefinition(name="somef", type=wf.TaskFieldDefinitionType.STRING)

        child_task_config = wf.ChildTaskConfiguration(
            task_definition=wf.TaskDefinitionRef(id="1", scope="somescope", code="somecode"),
            # Pyright thinks TaskFieldDefinition is unhashable
            child_task_fields={field: wf.FieldMapping(map_from=field)},  # pyright: ignore
        )

        ex = wf.ResultantChildTaskConfiguration(child_task_configuration=child_task_config)
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "taskDefinitionId": {"scope": "somescope", "code": "somecode"},
            "childTaskFields": {"somef": {"mapFrom": "somef"}},
        }

    @staticmethod
    def test_create_run_worker_action():
        worker = wf.WorkerResource(
            id="myworkerid", scope="a", code="b", display_name="b", worker_configuration=wf.Fail()
        )
        ex = wf.ActionDefinition(name="worker", action_details=wf.RunWorkerAction(worker=worker))

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "worker",
            "actionDetails": {"type": "RunWorker", "workerId": {"scope": "a", "code": "b"}},
        }
        ex = wf.ActionDefinition(
            name="worker",
            action_details=wf.RunWorkerAction(
                worker=worker, worker_as_at=datetime(year=2018, month=1, day=1, tzinfo=timezone.utc)
            ),
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "worker",
            "actionDetails": {
                "type": "RunWorker",
                "workerAsAt": "2018-01-01T00:00:00+00:00",
                "workerId": {"scope": "a", "code": "b"},
            },
        }
        child_task_config = wf.ChildTaskConfiguration(
            task_definition=wf.TaskDefinitionRef(id="1", scope="somescope", code="somecode"),
            child_task_fields={"somefield": wf.FieldMapping(map_from="somethingelse")},
        )
        ex = wf.ActionDefinition(
            name="someworker",
            action_details=wf.RunWorkerAction(
                worker=worker,
                worker_as_at="2018-01-01T00:00:00+00:00",
                worker_parameters={"somefield": wf.FieldMapping(map_from="somethingelse")},
                worker_status_triggers=wf.WorkerStatusTriggers(started="something"),
                child_task_configurations=[
                    wf.ResultantChildTaskConfiguration(
                        child_task_configuration=child_task_config, result_matching_pattern="something"
                    )
                ],
            ),
        )
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "someworker",
            "actionDetails": {
                "type": "RunWorker",
                "workerAsAt": "2018-01-01T00:00:00+00:00",
                "workerId": {"scope": "a", "code": "b"},
                "workerParameters": {"somefield": {"mapFrom": "somethingelse"}},
                "workerStatusTriggers": {"started": "something"},
                "childTaskConfigurations": [
                    {
                        "taskDefinitionId": {"scope": "somescope", "code": "somecode"},
                        "childTaskFields": {"somefield": {"mapFrom": "somethingelse"}},
                        "resultMatchingPattern": {"filter": "something"},
                    }
                ],
            },
        }

    @staticmethod
    def test_task_field_definition():
        ex = wf.TaskFieldDefinition(name="somefield", type=wf.TaskFieldDefinitionType.STRING)
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "somefield",
            "type": "String",
        }

        ex = wf.TaskFieldDefinition(
            name="somefield",
            type="DateTime",
            read_only_states=wf.ReadOnlyStates(
                state_type=wf.ReadOnlyStateType.ALL_STATES,
                selected_states=["something", wf.TaskStateDefinition(name="somestate")],
            ),
            value_constraints=wf.ValueConstraints(
                constraint_type=wf.ValueConstraintType.SUGGESTED,
                value_source_type="AcceptableValues",
                acceptable_values=["something1"],
            ),
        )
        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "somefield",
            "type": "DateTime",
            "readOnlyStates": {"stateType": "AllStates", "selectedStates": ["something", "somestate"]},
            "valueConstraints": {
                "constraintType": "Suggested",
                "valueSourceType": "AcceptableValues",
                "acceptableValues": ["something1"],
            },
        }

    @staticmethod
    def test_trigger_parent_task_action():
        ex = wf.ActionDefinition(
            name="trigger", action_details=wf.TriggerParentTaskAction(trigger="sometrigger")
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "trigger",
            "actionDetails": {"type": "TriggerParentTask", "trigger": "sometrigger"},
        }

        ex = wf.ActionDefinition(
            name="trigger",
            action_details=wf.TriggerParentTaskAction(
                trigger=wf.TriggerDefinition(name="sometrigger", type="External")
            ),
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "name": "trigger",
            "actionDetails": {"type": "TriggerParentTask", "trigger": "sometrigger"},
        }

    @staticmethod
    def test_task_transition_definition():
        ex = wf.TaskTransitionDefinition(from_state="state1", to_state="state2", trigger="sometrigger")

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "fromState": "state1",
            "toState": "state2",
            "trigger": "sometrigger",
        }
        ex = wf.TaskTransitionDefinition(
            from_state="state1", to_state="state2", trigger="sometrigger", action="someaction"
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "fromState": "state1",
            "toState": "state2",
            "trigger": "sometrigger",
            "action": "someaction",
        }

        ex = wf.TaskTransitionDefinition(
            from_state=wf.TaskStateDefinition(name="state1"),
            to_state=wf.TaskStateDefinition(name="state2"),
            trigger=wf.TriggerDefinition(name="sometrigger", type="External"),
            guard="someguard",
            action=wf.ActionDefinition(
                name="someaction", action_details=wf.TriggerParentTaskAction(trigger="sometrigger")
            ),
        )

        assert ex.model_dump(mode="json", by_alias=True, exclude_none=True) == {
            "fromState": "state1",
            "toState": "state2",
            "trigger": "sometrigger",
            "action": "someaction",
            "guard": "someguard",
        }


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeTaskDefinitionResource:
    base_url = TEST_BASE
    task_def_url = "/workflow/api/taskdefinitions"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @staticmethod
    def get_version(td: wf.TaskDefinitionResource) -> str:
        dump = td.model_dump(mode="json", by_alias=True, exclude_none=True)
        return hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()

    def test_create_task_definition(self, respx_mock):
        # GIVEN Resource does not exist
        state1 = wf.TaskStateDefinition(name="state1")
        state2 = wf.TaskStateDefinition(name="state2")
        tr = wf.TaskDefinitionResource(
            id="task1",
            scope="somescope",
            code="somecode",
            display_name="name",
            description="description",
            initial_state=wf.InitialState(name=state1),
            states=[state1],
            triggers=[wf.TriggerDefinition(name="sometrigger", type="External")],
            transitions=[
                wf.TaskTransitionDefinition(from_state=state1, to_state=state2, trigger="sometrigger")
            ],
            field_schema=[wf.TaskFieldDefinition(name="somefield", type="String")],
            actions=[
                wf.ActionDefinition(
                    name="someaction", action_details=wf.TriggerParentTaskAction(trigger="sometrigger")
                )
            ],
        )

        respx_mock.post(self.task_def_url).mock(
            side_effect=[
                httpx.Response(
                    status_code=201,
                    json={
                        "id": {"scope": "somescope", "code": "somecode"},
                        "version": {
                            "asAtCreated": "2024-07-09T08:40:56.7836370+00:00",
                            "asAtVersionNumber": "1",
                        },
                        "displayName": "name",
                        "description": "descriptions",
                        "states": [{"name": "state1"}, {"name": "state2"}],
                        "initialState": {"name": "state1", "requiredFields": []},
                        "fieldSchema": [{"name": "somefield", "type": "String"}],
                        "triggers": [{"name": "sometrigger", "type": "External"}],
                        "actions": [
                            {
                                "name": "someaction",
                                "actionDetails": {"type": "TriggerParentTask", "trigger": "sometrigger"},
                            }
                        ],
                    },
                )
            ]
        )

        # WHEN Task definition Resource is created
        new_state = tr.create(self.client)

        # THEN State returned as expected
        assert new_state == {
            "id": "task1",
            "scope": "somescope",
            "code": "somecode",
            "source_version": "8830ca77423f70290a811bb30c37527e129dfd1c80529d5c58b5ce66e1db4c85",
            "remote_version": "1",
        }

        request: httpx.Request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == self.task_def_url
        assert f"https://{request.url.host}" == self.base_url
        assert json.loads(request.content) == {
            "displayName": "name",
            "initialState": {"name": "state1", "requiredFields": []},
            "states": [{"name": "state1"}],
            "description": "description",
            "triggers": [{"name": "sometrigger", "trigger": {"type": "External"}}],
            "transitions": [{"fromState": "state1", "toState": "state2", "trigger": "sometrigger"}],
            "fieldSchema": [{"name": "somefield", "type": "String"}],
            "actions": [
                {
                    "name": "someaction",
                    "actionDetails": {"type": "TriggerParentTask", "trigger": "sometrigger"},
                }
            ],
            "id": {"scope": "somescope", "code": "somecode"},
        }

    @pytest.mark.parametrize("allow_none", [True, False])
    def test_update_no_change(self, respx_mock, allow_none: bool):
        class LocalFactory(ModelFactory[wf.TaskDefinitionResource]):
            # when set to False, factory will generate values for all fields
            # when set to True
            # factory will generate fields for required fields
            # and randomly assign either a value or None to optional fields
            # this test that update works for both cases
            __allow_none_optionals__ = allow_none

        task_def = LocalFactory.build()
        source_version = self.get_version(task_def)

        respx_mock.get(f"{self.task_def_url}/{task_def.scope}/{task_def.code}").mock(
            side_effect=[httpx.Response(200, json={"version": {"asAtVersionNumber": "1"}})]
        )

        old_state = SimpleNamespace(
            id=task_def.id,
            scope=task_def.scope,
            code=task_def.code,
            source_version=source_version,
            remote_version="1",
        )

        new_state = task_def.update(self.client, old_state)
        assert new_state is None

    def test_update_identifier(self, respx_mock):
        task_def = TaskDefFactory.build(scope="somethingelse", code="someothercode")
        old_state = SimpleNamespace(id="someid", scope="somescope", code="somecode")

        respx_mock.delete(f"{self.task_def_url}/{old_state.scope}/{old_state.code}").mock(
            side_effect=[httpx.Response(200, json={"asAt": "asAt deleted"})]
        )

        respx_mock.post(self.task_def_url).mock(
            side_effect=[httpx.Response(status_code=201, json={"version": {"asAtVersionNumber": "2"}})]
        )

        result = task_def.update(self.client, old_state)
        source_version = self.get_version(task_def)

        assert result == {
            "id": task_def.id,
            "scope": "somethingelse",
            "code": "someothercode",
            "source_version": source_version,
            "remote_version": "2",
        }

    def test_update_source_changes(self, respx_mock):
        existing_task_def = TaskDefFactory.build(
            scope="somethingelse", code="someothercode", display_name="something"
        )

        old_state = SimpleNamespace(
            id="someid",
            scope="somethingelse",
            code="someothercode",
            source_version=self.get_version(existing_task_def),
            remote_version="1",
        )

        new_task_def = TaskDefFactory.build(
            scope="somethingelse", code="someothercode", display_name="somethingelse"
        )

        respx_mock.get(f"{self.task_def_url}/{existing_task_def.scope}/{existing_task_def.code}").mock(
            side_effect=[httpx.Response(200, json={"version": {"asAtVersionNumber": "1"}})]
        )

        respx_mock.put(self.task_def_url + "/somethingelse/someothercode").mock(
            side_effect=[httpx.Response(status_code=201, json={"version": {"asAtVersionNumber": "2"}})]
        )

        new_state = new_task_def.update(self.client, old_state)

        assert new_state == {
            "id": new_task_def.id,
            "scope": "somethingelse",
            "code": "someothercode",
            "source_version": self.get_version(new_task_def),
            "remote_version": "2",
        }

        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == f"{self.task_def_url}/somethingelse/someothercode"
        assert json.loads(request.content) == new_task_def.model_dump(
            mode="json", by_alias=True, exclude_none=True
        )

    def test_update_remote_changes(self, respx_mock):
        existing_task_def = TaskDefFactory.build(scope="somethingelse", code="someothercode")

        old_state = SimpleNamespace(
            id="someid",
            scope="somethingelse",
            code="someothercode",
            source_version=self.get_version(existing_task_def),
            remote_version="1",
        )

        respx_mock.get(f"{self.task_def_url}/{existing_task_def.scope}/{existing_task_def.code}").mock(
            side_effect=[httpx.Response(200, json={"version": {"asAtVersionNumber": "2"}})]
        )

        respx_mock.put(self.task_def_url + "/somethingelse/someothercode").mock(
            side_effect=[httpx.Response(status_code=201, json={"version": {"asAtVersionNumber": "3"}})]
        )

        new_state = existing_task_def.update(self.client, old_state)

        assert new_state == {
            "id": existing_task_def.id,
            "scope": "somethingelse",
            "code": "someothercode",
            "source_version": self.get_version(existing_task_def),
            "remote_version": "3",
        }

        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == f"{self.task_def_url}/somethingelse/someothercode"
        assert json.loads(request.content) == existing_task_def.model_dump(
            mode="json", by_alias=True, exclude_none=True
        )

    def test_delete(self, respx_mock):
        respx_mock.delete(f"{self.task_def_url}/test/testcode").mock(
            side_effect=[httpx.Response(200, json={"asAt": "asAt deleted"})]
        )

        old_state = SimpleNamespace(id="tr1", scope="test", code="testcode")
        tr = TaskDefFactory.build()

        assert tr.delete(self.client, old_state) is None

        request = respx_mock.calls.last.request
        assert request.method == "DELETE"

    @staticmethod
    def test_deps():
        task_ref = TaskDefRefFactory.build()
        task_definition = TaskDefFactory.build()
        user = UserFactory.build()
        ct1 = wf.ChildTaskConfiguration(task_definition=task_ref)
        ct2 = wf.ChildTaskConfiguration(task_definition=task_definition)
        rt1 = wf.ResultantChildTaskConfiguration(child_task_configuration=ct1)
        rt2 = wf.ResultantChildTaskConfiguration(child_task_configuration=ct2)

        create_child_task_action1 = wf.CreateChildTasksAction(child_task_configurations=[ct1, ct2])
        assert create_child_task_action1.deps() == [task_ref, task_definition]
        create_child_task_action2 = wf.CreateChildTasksAction(child_task_configurations=[])
        assert create_child_task_action2.deps() == []

        # WHEN The worker name and descriptions change
        # But not the worker config
        wr = WorkerResourceFactory.build()
        job_ref = scheduler.JobRef(id="A", scope="jobscope", code="jobcode")
        wr.worker_configuration = wf.SchedulerJob(job=job_ref)
        worker_ref = wf.WorkerRef(id="wrkr", scope="PricingDemo", code="load_quotes")
        rwa = wf.RunWorkerAction(worker=wr, child_task_configurations=[rt1])
        rwa2 = wf.RunWorkerAction(worker=worker_ref, child_task_configurations=[rt1, rt2])

        assert rwa.deps() == [wr, task_ref]

        assert rwa2.deps() == [worker_ref, task_ref, task_definition]
        user_ref = identity.UserRef(id="a", login="b")

        action1 = wf.ActionDefinition(
            name="something", action_details=create_child_task_action1, run_as_user_id=user_ref
        )

        assert action1.deps() == [user_ref, task_ref, task_definition]

        action2 = wf.ActionDefinition(name="something", action_details=rwa2, run_as_user_id=user)

        assert action2.deps() == [user, worker_ref, task_ref, task_definition]

        action4 = wf.ActionDefinition(
            name="something", action_details=wf.TriggerParentTaskAction(trigger="something")
        )
        assert action4.deps() == []

        action3 = wf.ActionDefinition(name="something", action_details=rwa2, run_as_user_id=user)

        transition1 = wf.TaskTransitionDefinition(
            from_state="a", to_state="b", trigger="something", action=action3
        )
        assert transition1.deps() == [user, worker_ref, task_ref, task_definition]

        task_def = TaskDefFactory.build(transitions=[transition1], actions=[action3])
        target = task_def.deps()
        assert target == [user, worker_ref, task_ref, task_definition] + [
            user,
            worker_ref,
            task_ref,
            task_definition,
        ]


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeEventHandlerRef:
    base_url = TEST_BASE
    eh_url = "/workflow/api/eventhandlers"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get(f"{self.eh_url}/eh1/eh2").mock(
            side_effect=[httpx.Response(200, json={"id": {"scope": "w1", "code": "w2"}})]
        )

        sut = wf.EventHandlerRef(id="one", scope="eh1", code="eh2")
        # when we call attach
        sut.attach(self.client)
        # then a get request was made with the scope passed as a parameter
        req = respx_mock.calls.last.request
        assert req.url.path == f"{self.eh_url}/eh1/eh2"

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get(f"{self.eh_url}/eh1/eh2").mock(
            side_effect=[httpx.Response(404, json={"name": "EventHandlerNotFound"})]
        )
        sut = wf.EventHandlerRef(id="two", scope="eh1", code="eh2")

        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)

        assert "Event handler eh1/eh2 does not exist" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get(f"{self.eh_url}/eh1/eh2").mock(side_effect=[httpx.Response(400, json={})])
        client = self.client
        sut = wf.EventHandlerRef(id="three", scope="eh1", code="eh2")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeEventHandlerResource:
    base_url = TEST_BASE
    eh_url = "/workflow/api/eventhandlers"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def test_create_event_handler(self, respx_mock):
        # GIVEN Resource does not exist
        eh = wf.EventHandlerResource(
            id="id1",
            scope="eh1",
            code="eh2",
            display_name="new name",
            description="something",
            status=wf.EventStatus.INACTIVE,
            event_matching_pattern=wf.EventMatchingPattern(
                event_type="FileCreated", filter="body.filePath startswith '/somedomain/quotes'"
            ),
            run_as_user_id=wf.EventHandlerMapping(map_from="header.userId"),
            task_definition=TaskDefFactory.build(scope="a", code="b"),
            task_activity=wf.CreateNewTaskActivity(
                correlation_ids=[wf.EventHandlerMapping(set_to="int-test")],
                task_fields={"abc": wf.FieldMapping(map_from="body.filePath")},
                initial_trigger=wf.TriggerDefinition(name="abc", type="External"),
            ),
        )

        respx_mock.post(self.eh_url).respond(
            status_code=201, json={"version": {"asAtVersionNumber": "1"}}
        )

        new_state = eh.create(self.client)
        # THEN State returned as expected
        assert new_state == {
            "id": "id1",
            "scope": "eh1",
            "code": "eh2",
            "source_version": "699bfead44ca952993ea9fd271a602284dbf6e6818e295f94681e0d430804ced",
            "remote_version": "1",
        }

        request: httpx.Request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == self.eh_url
        assert f"https://{request.url.host}" == self.base_url
        assert json.loads(request.content) == {
            "description": "something",
            "displayName": "new name",
            "eventMatchingPattern": {
                "eventType": "FileCreated",
                "filter": "body.filePath startswith " "'/somedomain/quotes'",
            },
            "id": {"code": "eh2", "scope": "eh1"},
            "runAsUserId": {"mapFrom": "header.userId"},
            "status": "Inactive",
            "taskActivity": {
                "correlationIds": [{"setTo": "int-test"}],
                "initialTrigger": "abc",
                "taskFields": {"abc": {"mapFrom": "body.filePath"}},
                "type": "CreateNewTask",
            },
            "taskDefinitionId": {"code": "b", "scope": "a"},
        }

    def test_update_no_change(self, respx_mock):
        # GIVEN Resource exists
        respx_mock.get(f"{self.eh_url}/eh1/eh2").respond(
            status_code=201, json={"version": {"asAtVersionNumber": "1"}}
        )
        # AND No Change
        eh = EventHandlerFactory.build(scope="eh1", code="eh2")

        old_state = SimpleNamespace(
            id="wr1",
            scope="eh1",
            code="eh2",
            source_version=eh.__get_content_hash__(),  # pyright: ignore
            remote_version="1",
        )
        new_state = eh.update(self.client, old_state)
        # Then no actions
        assert new_state is None

    @pytest.mark.parametrize(
        "scope_and_code", [("eh1", "newcode"), ("newscope", "newcode"), ("newscope", "eh2")]
    )
    def test_update_identifier(self, respx_mock, scope_and_code):
        scope, code = scope_and_code
        # GIVEN An existing event handler
        # WHEN scope or code are changed
        # THEN Delete and recreate
        respx_mock.delete(f"{self.eh_url}/eh1/eh2").respond(200)
        respx_mock.post(self.eh_url).respond(201, json={"version": {"asAtVersionNumber": "2"}})

        old_state = SimpleNamespace(id="id1", scope="eh1", code="eh2")
        sut = EventHandlerFactory.build(scope=scope, code=code)

        new_state = sut.update(self.client, old_state)

        assert new_state
        assert new_state["scope"] == scope
        assert new_state["code"] == code
        assert new_state["remote_version"] == "2"

    def test_update_source(self, respx_mock):
        # GIVEN event handler exists

        # When local version changes

        # THEN Resource is updated
        existing_eh = EventHandlerFactory.build(
            id="someid", scope="somethingelse", code="someothercode", display_name="something"
        )
        dump = existing_eh.model_dump(
            mode="json", exclude_none=True, by_alias=True, exclude={"event_handler_id"}
        )
        source_version = hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
        old_state = SimpleNamespace(
            id="someid",
            scope="somethingelse",
            code="someothercode",
            source_version=source_version,
            remote_version="1",
        )

        new_eh = EventHandlerFactory.build(
            id="someid", scope="somethingelse", code="someothercode", display_name="somethingelse"
        )

        respx_mock.get(f"{self.eh_url}/{existing_eh.scope}/{existing_eh.code}").respond(
            status_code=200, json={"version": {"asAtVersionNumber": "1"}}
        )

        respx_mock.put(self.eh_url + "/somethingelse/someothercode").respond(
            status_code=201, json={"version": {"asAtVersionNumber": "2"}}
        )

        new_state = new_eh.update(self.client, old_state)
        dump = new_eh.model_dump(
            mode="json", exclude_none=True, by_alias=True, exclude={"event_handler_id"}
        )
        source_version = hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
        assert new_state == {
            "id": existing_eh.id,
            "scope": "somethingelse",
            "code": "someothercode",
            "source_version": source_version,
            "remote_version": "2",
        }

        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == f"{self.eh_url}/somethingelse/someothercode"
        assert json.loads(request.content) == dump

    def test_update_remote(self, respx_mock):
        # GIVEN event handler exists
        eh = EventHandlerFactory.build(
            id="someid", scope="somethingelse", code="someothercode", display_name="something"
        )
        dump = eh.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"event_handler_id"})
        source_version = hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
        old_state = SimpleNamespace(
            id="someid",
            scope="somethingelse",
            code="someothercode",
            source_version=source_version,
            remote_version="1",
        )
        # When remote version changes
        respx_mock.get(f"{self.eh_url}/{eh.scope}/{eh.code}").respond(
            status_code=200, json={"version": {"asAtVersionNumber": "2"}}
        )

        respx_mock.put(f"{self.eh_url}/{eh.scope}/{eh.code}").respond(
            status_code=201, json={"version": {"asAtVersionNumber": "3"}}
        )

        new_state = eh.update(self.client, old_state)
        # THEN Resource is updated
        assert new_state == {
            "id": eh.id,
            "scope": "somethingelse",
            "code": "someothercode",
            "source_version": source_version,
            "remote_version": "3",
        }

        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert json.loads(request.content) == dump

    def test_delete(self, respx_mock):
        respx_mock.delete(f"{self.eh_url}/test/testcode").respond(200)

        old_state = SimpleNamespace(id="tr1", scope="test", code="testcode")
        eh = EventHandlerFactory.build()

        assert eh.delete(self.client, old_state) is None

        request = respx_mock.calls.last.request
        assert request.method == "DELETE"

    @staticmethod
    def test_deps():
        eh = EventHandlerFactory.build(status="Active")

        assert len(eh.deps()) == 1
        assert isinstance(eh.deps()[0], (wf.TaskDefinitionResource, wf.TaskDefinitionRef))



from __future__ import annotations

import hashlib
from pathlib import PurePath, PurePosixPath
from typing import Optional, Union

import httpx
from pydantic import BaseModel, Field

from .resource_abc import Ref, Resource, register_resource


@register_resource()
class FolderRef(BaseModel, Ref):
    """Reference to a drive directory."""

    id: str = Field(exclude=True, init=True)
    drive_id: str = Field("/", exclude=True, init=False)
    folder_path: str

    def attach(self, client):
        if self.folder_path == "/":
            self.drive_id = "/"
            return
        p = PurePosixPath(self.folder_path)
        search = client.post("/drive/api/search/", json={"withPath": str(p.parent), "name": str(p.name)})
        values = search.json()["values"]
        if len(values) != 1:
            raise RuntimeError(
                "Expected to find exactly one match path for PathRef but found " + str(len(values))
            )
        self.drive_id = values[0]["id"]

    def path(self):
        return PurePosixPath(self.folder_path)


root = FolderRef(id="drive_root", folder_path="/")


@register_resource()
class FolderResource(BaseModel, Resource):
    """Resource describing a Drive folder

    Creates if it doesn't exist and updates if it has changed:
        - When name is changed, the reference for the folder will be maintained
        - When parent is changed, the folder will be moved under new parent

    Example
    -------
        >>> from fbnconfig import drive, Deployment
        >>> f1 = drive.FolderResource(id="base_folder", name="first_level", parent=drive.root)
        >>> f2 = drive.FolderResource(id="base_folder", name="second_level", parent=f1)
        >>> Deployment("myDeployment", [f1,f2])

    Notes
    -----
    If the folder is the parent of another resource, all dependents need to be deleted before this one

    Attributes
    ----------
    id : str
        resource identifier; this will be used in the log to reference the folder resource

    name : str
        folder name that will be displayed in Drive UI

    parent:  Union[FolderResource, RootFolder]
        Folder reference to the parent folder; if it is in root, use 'drive.root'
    """

    id: str = Field(exclude=True)
    drive_id: str | None = Field(None, exclude=True, init=False)
    name: str
    parent: Union[FolderResource, FolderRef]

    def read(self, client, old_state):
        pass  # not needed

    def create(self, client: httpx.Client):
        body = {"path": str(self.parent.path()), "name": self.name}
        res = client.request("POST", "/drive/api/folders", json=body)
        self.drive_id = res.json()["id"]
        return {
            "id": self.id,
            "driveId": self.drive_id,
            "name": self.name,
            "parentId": self.parent.drive_id,
        }

    def update(self, client: httpx.Client, old_state):
        self.drive_id = old_state.driveId
        if self.name != old_state.name or self.parent.drive_id != old_state.parentId:
            body = {"path": str(self.parent.path()), "name": self.name}
            client.request("PUT", "/drive/api/folders/" + old_state.driveId, json=body)
            return {
                "id": self.id,
                "driveId": self.drive_id,
                "name": self.name,
                "parentId": self.parent.drive_id,
            }
        return None

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", "/drive/api/folders/" + old_state.driveId)

    def deps(self):
        return [self.parent] if self.parent else []

    def path(self):
        return self.parent.path() / PurePosixPath(self.name)


@register_resource()
class FileResource(BaseModel, Resource):
    """Resource describing a Drive file.

    Creates if it doesn't exist and updates if content changes.

    Example
    -------
        >>> from fbnconfig import drive, Deployment
        >>> import pathlib
        >>> f1 = drive.FolderResource(id="base_folder", name="first_level", parent=drive.root)
        >>> content_path = pathlib.Path(__file__).parent.resolve() / pathlib.Path("myfile1.txt")
        >>> ff_with_path =
        >>> drive.FileResource(id="file1", folder=f1, name="myfile1.txt", content_path=content_path)
        >>> ff_with_content =
        >>> drive.FileResource(id="file2", folder=f1, name="myfile2.txt", content="Content of my file")
        >>> Deployment("myDeployment", [f1, ff_with_path, ff_with_content])

    Notes
    -----
    Can only supply a path to content or the content itself, but not both

    Attributes
    ----------
    id : str
      resource identifier; this will be used in the log to reference the file resource
    name : str
      file name that will be displayed in Drive UI
    content: Optional[Union[str, bytes]]
        file content
    content_path: Optional[PurePath]
        Path to the content of the file
    folder: FolderResource
      Folder reference to the parent folder; if it is in root, use 'drive.root'
    """

    id: str = Field(exclude=True)
    drive_id: str | None = Field(None, exclude=True, init=False)
    name: str
    content: Optional[Union[str, bytes]] = None
    content_path: Optional[PurePath] = None
    folder: FolderResource
    content_hash: str | None = Field(None, exclude=True, init=False)

    _drive_path = "/drive/api/files"

    def read(self, client, old_state) -> None:
        pass  # not needed

    def __init__(self, **options):
        super().__init__(**options)
        if self.content_path is not None:
            if self.content is not None:
                raise RuntimeError(
                    "Only one of content and content_path should be specified in FileResource"
                )
            if not self.content_path.is_absolute():
                raise RuntimeError("content_path should be an absolute path")
            with open(self.content_path, "rb") as ff:
                self.content = ff.read()
            self.content_hash = hashlib.sha256(self.content).hexdigest()
        elif self.content is not None:
            encoded = self.content.encode() if isinstance(self.content, str) else self.content
            self.content_hash = hashlib.sha256(encoded).hexdigest()

        else:
            raise RuntimeError("Either content or content_path should be specified in FileResource")

    def create(self, client):
        path = str(self.folder.path())
        res = client.request(
            "POST",
            self._drive_path,
            headers={
                "x-lusid-drive-filename": self.name,
                "x-lusid-drive-path": path,
                "content-type": "application/octet-stream",
            },
            content=self.content,
        )
        self.drive_id = res.json()["id"]
        return {
            "id": self.id,
            "driveId": self.drive_id,
            "name": self.name,
            "parentId": self.folder.drive_id,
            "content_hash": self.content_hash,
        }

    def update(self, client, old_state):
        self.drive_id = old_state.driveId
        if (
            self.name == old_state.name
            and self.folder.drive_id == old_state.parentId
            and self.content_hash == old_state.content_hash
        ):
            return None

        if self.content_hash != old_state.content_hash:
            assert self.drive_id is not None
            client.request(
                "put",
                "/drive/api/files/" + self.drive_id + "/contents",
                headers={"content-type": "application/octet-stream"},
                content=self.content,
            )
        if self.name != old_state.name or self.folder.drive_id != old_state.parentId:
            json = {"path": str(self.folder.path()), "name": self.name}
            client.request("PUT", self._drive_path + "/" + old_state.driveId, json=json)
        return {
            "id": self.id,
            "driveId": self.drive_id,
            "name": self.name,
            "parentId": self.folder.drive_id,
            "content_hash": self.content_hash,
        }

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", "/drive/api/files/" + old_state.driveId)

    def deps(self):
        return [self.folder]

    def path(self):
        return self.folder.path() / PurePosixPath(self.name)



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import configuration

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeSetRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_when_exists(self, respx_mock):
        # given a set exists
        respx_mock.get("/configuration/api/sets/personal/scope1/code1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        # when we attach
        sut = configuration.SetRef(
            id="abc", scope="scope1", code="code1", type=configuration.SetType.PERSONAL
        )
        sut.attach(client)
        # then all is good an no exception

    def test_attach_when_not_exists(self, respx_mock):
        # given a set exists
        respx_mock.get("/configuration/api/sets/personal/scope1/code1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        # when we attach
        sut = configuration.SetRef(
            id="abc", scope="scope1", code="code1", type=configuration.SetType.PERSONAL
        )
        # a runtime error is raised
        with pytest.raises(RuntimeError):
            sut.attach(client)

    def test_attach_when_http_error(self, respx_mock):
        # given a server which returns a 500
        respx_mock.get("/configuration/api/sets/personal/scope1/code1").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        # when we attach
        sut = configuration.SetRef(
            id="abc", scope="scope1", code="code1", type=configuration.SetType.PERSONAL
        )
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeSetResource:
    client = httpx.Client(base_url=TEST_BASE)

    def test_create(self, respx_mock):
        respx_mock.post("/configuration/api/sets").mock(return_value=httpx.Response(200, json={}))
        # given a desired set
        sut = configuration.SetResource(
            id="xyz",
            scope="scope-a",
            code="code-a",
            type=configuration.SetType.SHARED,
            description="blah",
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned
        assert state == {"scope": "scope-a", "code": "code-a", "type": "shared"}
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/configuration/api/sets"
        assert json.loads(request.content) == {
            "id": {"scope": "scope-a", "code": "code-a"},
            "type": "shared",
            "description": "blah",
        }

    def test_update_with_no_changes(self, respx_mock):
        # given an existing set
        respx_mock.get("/configuration/api/sets/personal/scope-b/code-b").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {"scope": "scope-b", "code": "code-b"},
                    "type": "personal",
                    "description": "nice set!",
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "items": [],
                    "links": {},
                },
            )
        )
        # and a desired set
        sut = configuration.SetResource(
            id="xyz",
            scope="scope-b",
            code="code-b",
            type=configuration.SetType.PERSONAL,
            description="nice set!",
        )
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="personal")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and a read was made but no PUT

    def test_update_with_modified_description(self, respx_mock):
        # given an existing set
        respx_mock.get("/configuration/api/sets/personal/scope-b/code-b").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {"scope": "scope-b", "code": "code-b"},
                    "type": "personal",
                    "description": "nice set!",
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "items": [],
                    "links": {},
                },
            )
        )
        respx_mock.put("/configuration/api/sets/personal/scope-b/code-b").mock(
            return_value=httpx.Response(200, json={})
        )
        # and a desired set
        sut = configuration.SetResource(
            id="xyz",
            scope="scope-b",
            code="code-b",
            type=configuration.SetType.PERSONAL,
            description="great party Jeff!",
        )
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="personal")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the existing state is returned
        assert state == {"scope": "scope-b", "code": "code-b", "type": "personal"}
        # and the put is sent
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/configuration/api/sets/personal/scope-b/code-b"
        assert json.loads(request.content) == {"description": "great party Jeff!"}

    def test_update_with_change_scope_should_throw(self, respx_mock):
        sut = configuration.SetResource(
            id="xyz",
            scope="new-scope",
            code="code-b",
            type=configuration.SetType.PERSONAL,
            description="great party Jeff!",
        )
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="personal")
        # when we update it throws
        with pytest.raises(RuntimeError):
            sut.update(self.client, old_state)

    def test_update_with_change_code_should_throw(self, respx_mock):
        sut = configuration.SetResource(
            id="xyz",
            scope="scope-b",
            code="new-code",
            type=configuration.SetType.PERSONAL,
            description="great party Jeff!",
        )
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="personal")
        # when we update it throws
        with pytest.raises(RuntimeError):
            sut.update(self.client, old_state)

    def test_update_with_change_type_should_throw(self, respx_mock):
        sut = configuration.SetResource(
            id="xyz",
            scope="scope-b",
            code="code-b",
            type=configuration.SetType.SHARED,
            description="great party Jeff!",
        )
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="personal")
        # when we update it throws
        with pytest.raises(RuntimeError):
            sut.update(self.client, old_state)

    def test_delete(self, respx_mock):
        respx_mock.delete("/configuration/api/sets/shared/scope-b/code-b").mock(
            return_value=httpx.Response(200, json={})
        )
        # given a set that exists in the remnte
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="shared")
        # when we delete it
        configuration.SetResource.delete(self.client, old_state)
        # then a delete request is sent with the roleId
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/configuration/api/sets/shared/scope-b/code-b"

    def test_deps(self):
        sut = configuration.SetResource(
            id="xyz",
            scope="scope-b",
            code="code-b",
            type=configuration.SetType.SHARED,
            description="great party Jeff!",
        )
        # it's deps are empty
        assert sut.deps() == []


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeItemRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def existing_set(self):
        return configuration.SetRef(
            id="abc", scope="scope1", code="code1", type=configuration.SetType.PERSONAL
        )

    def test_attach_when_exists(self, respx_mock, existing_set):
        # given an item in the remote
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(200, json={"ref": "config://xyz"})
        )
        client = self.client
        # when we attach
        sut = configuration.ItemRef(id="111", set=existing_set, key="key1")
        sut.attach(client)
        # then all is good no exception and the ref is set
        assert sut.ref == "config://xyz"

    def test_attach_when_not_exists(self, respx_mock, existing_set):
        # given no matching remote
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        # when we attach
        sut = configuration.ItemRef(id="111", set=existing_set, key="key1")
        # it throws
        with pytest.raises(RuntimeError):
            sut.attach(client)

    def test_attach_when_http_error(self, respx_mock, existing_set):
        # given a 500
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        # when we attach
        sut = configuration.ItemRef(id="111", set=existing_set, key="key1")
        # a http error
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeItemResource:
    client = httpx.Client(base_url=TEST_BASE)

    @pytest.fixture
    def existing_set(self):
        return configuration.SetRef(
            id="set1", scope="scope1", code="code1", type=configuration.SetType.PERSONAL
        )

    def test_create(self, respx_mock, existing_set):
        respx_mock.post("/configuration/api/sets/personal/scope1/code1/items").mock(
            return_value=httpx.Response(200, json={"items": [{"key": "key1", "ref": "config://1234"}]})
        )
        # given a desired item
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something",
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned
        assert state == {"scope": "scope1", "code": "code1", "type": "personal", "key": "key1"}
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/configuration/api/sets/personal/scope1/code1/items"
        assert json.loads(request.content) == {
            "key": "key1",
            "value": "secret-value",
            "valueType": "text",
            "isSecret": False,
            "description": "something",
            "blockReveal": False,
        }
        # and the ref value has been set
        assert sut.ref == "config://1234"

    def test_update_with_no_change(self, respx_mock, existing_set):
        # given a matching item aready exists in the remote
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "key": "key1",
                    "value": "secret-value",
                    "valueType": "text",
                    "isSecret": False,
                    "description": "something",
                    "blockReveal": False,
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "ref": "config/123",
                },
            )
        )
        # and a desired item which matches the remote
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something",
        )
        old_state = SimpleNamespace(scope="scope1", code="code1", type="personal", key="key1")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and no put request was made
        # and the ref value has been set
        assert sut.ref == "config/123"

    def test_update_with_modified_valuetype(self, respx_mock, existing_set):
        # given an item already exists in the remote with a value of false
        # but valueType of text
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "key": "key1",
                    "value": "secret-value",
                    "valueType": "text",
                    "isSecret": False,
                    "description": "something",
                    "blockReveal": False,
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "ref": "config/123",
                },
            )
        )
        respx_mock.post("/configuration/api/sets/personal/scope1/code1/items").mock(
            return_value=httpx.Response(200, json={"items": [{"key": "key1", "ref": "config://1234"}]})
        )
        respx_mock.delete("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(200, json={"ref": "config/123"})
        )
        # and a desired item which changes the value type to boolean
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="false",
            value_type=configuration.ValueType.BOOLEAN,
            is_secret=False,
            description="something new",
        )
        old_state = SimpleNamespace(scope="scope1", code="code1", type="personal", key="key1")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the existing state is returned
        assert state == {"scope": "scope1", "code": "code1", "type": "personal", "key": "key1"}
        # and a delete and post are done to recreate with a different valuetype
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/configuration/api/sets/personal/scope1/code1/items"
        assert json.loads(request.content) == {
            "valueType": "boolean",
            "value": "false",
            "description": "something new",
            "blockReveal": False,
            "isSecret": False,
            "key": "key1",
        }

    def test_update_with_modified_description(self, respx_mock, existing_set):
        # given a matching item already exists in the remote
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "key": "key1",
                    "value": "secret-value",
                    "valueType": "text",
                    "isSecret": False,
                    "description": "something",
                    "blockReveal": False,
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "ref": "config/123",
                },
            )
        )
        respx_mock.put("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(200, json={"ref": "config/123"})
        )
        # and a desired item which matches the remote
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something new",
        )
        old_state = SimpleNamespace(scope="scope1", code="code1", type="personal", key="key1")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the existing state is returned
        assert state == {"scope": "scope1", "code": "code1", "type": "personal", "key": "key1"}
        # and the put is sent
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/configuration/api/sets/personal/scope1/code1/items/key1"
        assert json.loads(request.content) == {
            "value": "secret-value",
            "description": "something new",
            "blockReveal": False,
        }

    def test_update_with_modified_block_reveal(self, respx_mock, existing_set):
        # given a matching item already exists in the remote
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "key": "key1",
                    "value": "secret-value",
                    "valueType": "text",
                    "isSecret": False,
                    "description": "something",
                    "blockReveal": False,
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "ref": "config/123",
                },
            )
        )
        respx_mock.put("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(200, json={"ref": "config/123"})
        )
        # and a desired item which matches the remote
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something new",
            block_reveal=True,
        )
        old_state = SimpleNamespace(scope="scope1", code="code1", type="personal", key="key1")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the existing state is returned
        assert state == {"scope": "scope1", "code": "code1", "type": "personal", "key": "key1"}
        # and the put is sent
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/configuration/api/sets/personal/scope1/code1/items/key1"
        assert json.loads(request.content) == {
            "value": "secret-value",
            "description": "something new",
            "blockReveal": True,
        }

    def test_update_with_modified_location_recreates(self, respx_mock, existing_set):
        respx_mock.post("/configuration/api/sets/personal/scope1/code1/items").mock(
            return_value=httpx.Response(200, json={"items": [{"key": "key1", "ref": "config://1234"}]})
        )
        respx_mock.delete("/configuration/api/sets/personal/scope1/oldcode/items/key1").mock(
            return_value=httpx.Response(200, json={"ref": "config/123"})
        )
        # and a desired item which matches the remote
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something new",
        )
        old_state = SimpleNamespace(scope="scope1", code="oldcode", type="personal", key="key1")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the existing state is returned
        assert state == {"scope": "scope1", "code": "code1", "type": "personal", "key": "key1"}
        # a delete and a post are made

    def test_update_with_modified_details_recreates(self, respx_mock, existing_set):
        # given an existing item which is secret
        respx_mock.get("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "key": "key1",
                    "value": "secret-value",
                    "valueType": "text",
                    "isSecret": True,
                    "description": "something",
                    "createdAt": "2023-01-02",
                    "createdBy": "sarah",
                    "lastModifiedAt": "",
                    "lastModifiedBy": "",
                    "ref": "config/123",
                },
            )
        )
        respx_mock.post("/configuration/api/sets/personal/scope1/code1/items").mock(
            return_value=httpx.Response(200, json={"items": [{"key": "key1", "ref": "config://1234"}]})
        )
        respx_mock.delete("/configuration/api/sets/personal/scope1/code1/items/key1").mock(
            return_value=httpx.Response(200, json={"ref": "config/123"})
        )
        # and a desired item which is not
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something new",
        )
        old_state = SimpleNamespace(scope="scope1", code="code1", type="personal", key="key1")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the existing state is returned
        assert state == {"scope": "scope1", "code": "code1", "type": "personal", "key": "key1"}
        # a delete and a post are made

    def test_delete(self, respx_mock):
        respx_mock.delete("/configuration/api/sets/shared/scope-b/code-b/items/key1").mock(
            return_value=httpx.Response(200, json={})
        )
        # given a set that exists in the remnte
        old_state = SimpleNamespace(scope="scope-b", code="code-b", type="shared", key="key1")
        # when we delete it
        configuration.ItemResource.delete(self.client, old_state)
        # then a delete request is sent with the roleId
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/configuration/api/sets/shared/scope-b/code-b/items/key1"

    @staticmethod
    def test_deps(existing_set):
        # and a desired item which matches the remote
        sut = configuration.ItemResource(
            id="xyz",
            set=existing_set,
            key="key1",
            value="secret-value",
            value_type=configuration.ValueType.TEXT,
            is_secret=False,
            description="something new",
        )
        # it depends on the set
        assert sut.deps() == [existing_set]


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeSystemConfig:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create_given_config_does_not_exist_then_fail(self, respx_mock):
        (
            respx_mock.put("/configuration/api/sets/shared/system/somecode/items/shouldnotexist").mock(
                return_value=httpx.Response(
                    404,
                    json={
                        "name": "ConfigurationNotFound",
                        "code": 710,
                        "title": "Configuration item not found",
                        "detail": "Server message for not found",
                        "status": 404,
                    },
                )
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="somecode",
            key="shouldnotexist",
            value="whatever",
            description="something",
            default_value="something",
        )

        from httpx import HTTPStatusError

        with pytest.raises(HTTPStatusError) as ex:
            sut.create(self.client)
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert "404 Not Found" in ex.exconly(tryshort=True)

    def test_create_given_config_exists_then_updates(self, respx_mock):
        respx_mock.put(
            "/configuration/api/sets/shared/system/TransactionBooking/items/ValidateInstruments"
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "key": "ValidateInstruments",
                    "value": "False",
                    "valueType": "boolean",
                    "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                },
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="TransactionBooking",
            key="ValidateInstruments",
            value="False",
            description="something",
            default_value=None,
        )

        state = sut.create(self.client)
        assert state == {
            "code": "TransactionBooking",
            "key": "ValidateInstruments",
            "default_value": None,
        }

    def test_read_given_config_exists(self, respx_mock):
        (
            respx_mock.get(
                "/configuration/api/sets/system/TransactionBooking/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "key": "ValidateInstruments",
                                "value": "True",
                                "valueType": "boolean",
                                "isSecret": "false",
                                "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                            }
                        ]
                    },
                )
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="TransactionBooking",
            key="ValidateInstruments",
            value="False",
            description="something",
            default_value="something",
        )
        old_state = {"code": "TransactionBooking", "key": "ValidateInstruments"}
        sut.read(self.client, SimpleNamespace(**old_state))

    def test_update_given_no_change_then_no_calls(self, respx_mock):
        (
            respx_mock.get(
                "/configuration/api/sets/system/TransactionBooking/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "key": "ValidateInstruments",
                                "value": "True",
                                "valueType": "boolean",
                                "isSecret": "false",
                                "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                                "description": "something",
                                "blockReveal": False,
                            }
                        ]
                    },
                )
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="TransactionBooking",
            key="ValidateInstruments",
            value="True",
            description="something",
            default_value="something",
        )

        old_state = {"code": "TransactionBooking", "key": "ValidateInstruments"}
        state = sut.update(self.client, SimpleNamespace(**old_state))
        assert state is None

    def test_update_given_value_and_description_change(self, respx_mock):
        (
            respx_mock.get(
                "/configuration/api/sets/system/TransactionBooking/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "values": [
                            {
                                "key": "ValidateInstruments",
                                "value": "True",
                                "valueType": "boolean",
                                "isSecret": "false",
                                "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                                "description": "something else",
                                "blockReveal": False,
                            }
                        ]
                    },
                )
            )
        )

        (
            respx_mock.put(
                "/configuration/api/sets/shared/system/TransactionBooking/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "key": "ValidateInstruments",
                        "value": "False",
                        "valueType": "boolean",
                        "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                        "description": "something else",
                        "blockReveal": False,
                    },
                )
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="TransactionBooking",
            key="ValidateInstruments",
            value="False",
            description="something",
            default_value=None,
            block_reveal=True,
        )

        old_state = {"code": "TransactionBooking", "key": "ValidateInstruments", "default_value": None}
        state = sut.update(self.client, SimpleNamespace(**old_state))
        assert state == old_state
        request = respx_mock.calls.last.request
        assert json.loads(request.content) == {
            "value": "False",
            "description": "something",
            "blockReveal": True,
        }

    def test_update_with_modified_key_deletes_and_recreates(self, respx_mock):
        (
            respx_mock.put(
                "/configuration/api/sets/shared/system/TransactionBooking/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "key": "ValidateInstruments",
                        "value": "False",
                        "valueType": "boolean",
                        "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                        "description": "something else",
                    },
                )
            )
        )

        (
            respx_mock.put(
                "/configuration/api/sets/shared/system/TransactionBooking/items/SomeOtherKey"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "key": "SomeOtherKey",
                        "value": "True",
                        "valueType": "boolean",
                        "ref": "config://shared/system/TransactionBooking/SomeOtherKey",
                        "description": "something",
                    },
                )
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="TransactionBooking",
            key="SomeOtherKey",
            value="True",
            description="something",
            default_value=None,
        )

        old_state = {
            "scope": "system",
            "code": "TransactionBooking",
            "type": "shared",
            "key": "ValidateInstruments",
            "default_value": "False",
        }

        state = sut.update(self.client, SimpleNamespace(**old_state))
        assert state is not None
        assert state["key"] == "SomeOtherKey"
        assert state["code"] == "TransactionBooking"

    def test_update_with_modified_code_deletes_and_recreates(self, respx_mock):
        (
            respx_mock.put(
                "/configuration/api/sets/shared/system/TransactionBooking/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "key": "ValidateInstruments",
                        "value": "False",
                        "valueType": "boolean",
                        "ref": "config://shared/system/TransactionBooking/ValidateInstruments",
                        "description": "something else",
                    },
                )
            )
        )

        (
            respx_mock.put(
                "/configuration/api/sets/shared/system/SomeOtherCode/items/ValidateInstruments"
            ).mock(
                return_value=httpx.Response(
                    200,
                    json={
                        "key": "ValidateInstruments",
                        "value": "True",
                        "valueType": "boolean",
                        "ref": "config://shared/system/SomeOtherCode/SomeOtherKey",
                        "description": "something",
                    },
                )
            )
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="SomeOtherCode",
            key="ValidateInstruments",
            value="True",
            description="something",
            default_value=None,
        )

        old_state = {
            "code": "TransactionBooking",
            "key": "ValidateInstruments",
            "default_value": "False",
        }

        state = sut.update(self.client, SimpleNamespace(**old_state))
        assert state is not None
        assert state["key"] == "ValidateInstruments"
        assert state["code"] == "SomeOtherCode"

    def test_delete(self):
        sut = configuration.SystemConfigResource(
            id="pass",
            code="TransactionBooking",
            key="Something",
            value="False",
            description="some",
            default_value=None,
        )
        # nothing should happen
        sut.delete(self.client, SimpleNamespace(**{"default_value": None}))

    @staticmethod
    def test_deps():
        sut = configuration.SystemConfigResource(
            id="pass",
            code="somecode",
            key="Something",
            value="False",
            description="some",
            default_value=None,
        )
        assert sut.deps() == []

    def test_delete_given_default_value_then_updates(self, respx_mock):
        (
            respx_mock.put(
                "/configuration/api/sets/shared/system/TransactionBooking/items/ValidateInstruments"
            ).mock(return_value=httpx.Response(200, json={"something": "something"}))
        )

        sut = configuration.SystemConfigResource(
            id="txnbooking",
            code="TransactionBooking",
            key="ValidateInstruments",
            value="True",
            description="something",
            default_value="False",
        )
        old_state = {
            "code": "TransactionBooking",
            "key": "ValidateInstruments",
            "default_value": "False",
        }
        sut.delete(self.client, SimpleNamespace(**old_state))

        request = respx_mock.calls.last.request
        assert json.loads(request.content) == {"value": "False"}



from fbnconfig import Deployment, identity

"""
An example configuration for defining Identity related entities.
The script configures the following entities:
- Role
- User

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01653/
https://support.lusid.com/knowledgebase/article/KA-01794/
"""


def configure(env):
    admin = identity.IdentityRoleRef(id="admin-ref", name="instrument-administrator")
    operator = identity.IdentityRoleResource(id="role1", description="scheduler-operators", name="role1")

    # ! WARNING: this will create a user in your domain, please remove the following UserResource if you
    # do not wish for that to happen.
    user = identity.UserResource(
        id="user1",
        first_name="Example",
        last_name="User",
        email_address="example_user@example.org",
        login="example_service_user@example.com",
        type=identity.UserType.SERVICE,
    )
    assignment1 = identity.RoleAssignment(id="operator", user=user, role=operator)
    assignment2 = identity.RoleAssignment(id="admin", user=user, role=admin)
    return Deployment("identity_example", [assignment1, assignment2, user, operator])



from fbnconfig import Deployment, scheduler
from fbnconfig import workflows as wf


def configure(host_vars) -> Deployment:
    deployment_name = getattr(host_vars, "name", "workflows_test")
    tag = "v5.0"
    i = scheduler.ImageResource(
        id="img1",
        source_image="harbor.finbourne.com/ceng/fbnconfig-pipeline:0.1",
        dest_name="beany",
        dest_tag=tag,
    )
    job = scheduler.JobResource(
        id="job1",
        scope=deployment_name,
        code="jobwf",
        image=i,
        name="wfjob",
        description="something nice",
        min_cpu="1",
        max_cpu="2",
        argument_definitions={},
    )

    worker = wf.WorkerResource(
        id="wr3-example",
        scope=deployment_name,
        code="wr3",
        display_name="I am worker3",
        worker_configuration=wf.SchedulerJob(job=job),
    )

    pending_state = wf.TaskStateDefinition(name="pending")
    in_progress_state = wf.TaskStateDefinition(name="inprogress")
    end_state = wf.TaskStateDefinition(name="end")
    start_trigger = wf.TriggerDefinition(name="start", type="External")
    end_trigger = wf.TriggerDefinition(name="end", type="External")

    some_field = wf.TaskFieldDefinition(name="imafield", type=wf.TaskFieldDefinitionType.STRING)
    do_something_action = wf.ActionDefinition(
        name="start-something-worker",
        action_details=wf.RunWorkerAction(
            worker=worker,
            worker_parameters={},
            worker_status_triggers=wf.WorkerStatusTriggers(completed_with_results=end_trigger),
        ),
    )

    task_def = wf.TaskDefinitionResource(
        id="integrationtest-task-definition",
        scope=deployment_name,
        code="DoSomething",
        display_name="Does something",
        description="Task description",
        states=[pending_state, in_progress_state, end_state],
        field_schema=[some_field],
        initial_state=wf.InitialState(name=pending_state),
        triggers=[start_trigger, end_trigger],
        transitions=[
            wf.TaskTransitionDefinition(
                from_state=pending_state,
                to_state=in_progress_state,
                trigger=start_trigger,
                action=do_something_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_progress_state, to_state=end_state, trigger=end_trigger
            ),
        ],
        actions=[do_something_action],
    )

    event_handler = wf.EventHandlerResource(
        id="integrationtest-eventhandler",
        scope=deployment_name,
        code="testeventhandler",
        display_name="new name",
        description="something",
        status=wf.EventStatus.INACTIVE,
        event_matching_pattern=wf.EventMatchingPattern(
            event_type="FileCreated",
            filter="body.filePath startswith '/somedomain/quotes'"),
        run_as_user_id=wf.EventHandlerMapping(
            map_from="header.userId"
        ),
        task_definition=task_def,
        task_activity=wf.CreateNewTaskActivity(
            correlation_ids=[wf.EventHandlerMapping(set_to="int-test")],
            task_fields={
                some_field: wf.FieldMapping(map_from="body.filePath")  # pyright: ignore
            },
            initial_trigger=start_trigger
        )
    )

    return Deployment(deployment_name, [event_handler])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deploy(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import os

from fbnconfig import Deployment, drive, lumi
from fbnconfig import workflows as wf


def configure(host_vars) -> Deployment:
    deployment_name = getattr(host_vars, "name", "fbnconfig_quality_control_wf")
    file_name = "exampledata.xlsx"

    base_folder = drive.FolderResource(
        id="base_folder", name=f"fbnconfig-{deployment_name}", parent=drive.root
    )
    excel_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "data", file_name)
    deployment_folder = drive.FolderResource(id="sub_folder", name=deployment_name, parent=base_folder)

    spreadsheet = drive.FileResource(
        id="xslx", folder=deployment_folder, name=file_name, content_path=excel_path
    )

    import_from_excel_view = lumi.ViewResource(
        id="import-from-excel-view",
        provider=f"Views.fbnconfig.import_excel_{deployment_name}",
        description="I am a view",
        sql="""
            @@filename = select #PARAMETERVALUE(filename);
            @@quote_scope = select #PARAMETERVALUE(quote_scope);

            -- Load data from Excel
            @inst_data = use Drive.Excel with @@filename
                --file={@@filename}
                --worksheet=instrument
            enduse;

            @quote_data = use Drive.Excel with @@filename
                --file={@@filename}
                --worksheet=price_time_series
            enduse;

            -- Transform quote data
            @quotes_for_upload =
                select
                    'ClientInternal' as InstrumentIdType,
                    instrument_id as InstrumentId,
                    @@quote_scope as QuoteScope,
                    'Price' as QuoteType,
                    'Lusid' as Provider,
                    'Mid' as Field,
                    price_date as QuoteEffectiveAt,
                    price as Value,
                    ccy as Unit
                from @quote_data;

            -- Transform instrument data
            @equity_instruments =
                select
                    inst_id as ClientInternal,
                    name as DisplayName,
                    ccy as DomCcy,
                    @@quote_scope as Scope
                from @inst_data;


            -- Return quotes in view
            select * from @quotes_for_upload;

            -- Create instruments if not Active
            select *
            from Lusid.Instrument.Equity.Writer
            where ToWrite = @equity_instruments;

            -- Upload quotes into LUSID
            select *
            from Lusid.Instrument.Quote.Writer
            where ToWrite = @quotes_for_upload
        """,
        parameters=[
            lumi.Parameter(
                name="filename",
                value=f"{spreadsheet.path()}",
                set_as_default_value=False,
                tooltip="Drive filepath",
                type=lumi.ParameterType.Text,
            ),
            lumi.Parameter(
                name="quote_scope",
                value=deployment_name,
                set_as_default_value=False,
                tooltip="Scope to load quotes in",
                type=lumi.ParameterType.Text,
            ),
        ],
        dependencies=[spreadsheet],
    )

    reasonable_value_check_view = lumi.ViewResource(
        id="reasonable-value-check",
        provider=f"Views.fbnconfig.reasonable_value_{deployment_name}",
        sql="""
            @@quote_scope = select #PARAMETERVALUE(quote_scope);
            -- Collect quotes for all instruments
            @quotes_data = select *
                from Lusid.Instrument.Quote
                where QuoteScope = @@quote_scope
                    and InstrumentIdType = 'ClientInternal'
                    and QuoteType = 'Price';

            -- Collect instrument static
            @instrument_data = select
                ClientInternal,
                DisplayName
                from Lusid.Instrument.Equity
                where @@quote_scope = Scope
                    and State = 'Active';

            -- Generate time series
            @price_ts = select
                ClientInternal,
                DisplayName,
                QuoteEffectiveAt as [PriceDate],
                Unit as [Currency],
                Value as [Price],
                Field
                from @instrument_data i
                join @quotes_data q on (i.ClientInternal = q.InstrumentId);

            -- Run reasonable value check for each quote
            @result = select
                PriceDate,
                @@quote_scope as QuoteScope,
                ClientInternal,
                DisplayName,
                Price,
                Field,
                case
                    when Price >= 1000 then 'Unreasonably Large Value'
                    when Price <= 1 then 'Unreasonably Small Value'
                    else 'OK'
                end as Result
                from @price_ts
                where not Result = 'OK';

            select #SELECT {
                {PriceDate~DateTime : PriceDate},
                {QuoteScope~Text : QuoteScope},
                {ClientInternal~Text : ClientInternal},
                {DisplayName~Text : DisplayName},
                {Price~Decimal : Price},
                {Field~Text : Field},
                {Result~Text : Result}
                }
                from @result;
        """,
        description="Ensure a quote's price is a sensible value (between 1 and 1000)",
        parameters=[
            lumi.Parameter(
                name="quote_scope",
                value=deployment_name,
                set_as_default_value=True,
                tooltip="Scope to check quotes in",
                type=lumi.ParameterType.Text,
            )
        ],
    )

    iqr_outlier_view = lumi.ViewResource(
        id="iqr_outlier",
        provider=f"Views.fbnconfig.iqr_outlier_{deployment_name}",
        sql="""
            @@quote_scope = select #PARAMETERVALUE(quote_scope);
            -- Collect quotes for all instruments
            @quotes_data = select *
                from Lusid.Instrument.Quote
                where QuoteScope = @@quote_scope
                    and InstrumentIdType = 'ClientInternal'
                    and QuoteType = 'Price';

            -- Collect instrument static
            @instrument_data = select
                ClientInternal,
                DisplayName
                from Lusid.Instrument.Equity
                where @@quote_scope = Scope
                    and State = 'Active';

            -- Generate time series
            @price_ts = select
                ClientInternal,
                DisplayName,
                Field,
                QuoteEffectiveAt as [PriceDate],
                Unit as [Currency],
                Value as [Price]
                from @instrument_data i
                join @quotes_data q on (i.ClientInternal = q.InstrumentId);
             -- Run IQR checks for each instrument
            @iqr_data = select
                ClientInternal,
                interquartile_range(price) * (1.5) as [iqr_x1_5],
                quantile(price, 0.25) as [q1],
                quantile(price, 0.75) as [q3]
                from @price_ts
                group by ClientInternal;

            -- Join the IQR data with the time series and identify outliers
            @result = select
                p.PriceDate as PriceDate,
                @@quote_scope as QuoteScope,
                p.ClientInternal,
                p.DisplayName as DisplayName,
                i.q1,
                i.q3,
                (i.q3 + i.iqr_x1_5) as [UpperLimit],
                (i.q1 - i.iqr_x1_5) as [LowerLimit],
                p.Price as Price,
                Field,
                case when p.Price not between (i.q1 - i.iqr_x1_5) and (i.q3 + i.iqr_x1_5)
                    then 'IQR Outlier'
                    else 'OK'
                end as Result
                from @price_ts p
                join @iqr_data i on p.ClientInternal = i.ClientInternal
                where not Result = 'OK';

                select #SELECT {
                    {PriceDate~DateTime : PriceDate},
                    {QuoteScope~Text : QuoteScope},
                    {ClientInternal~Text : ClientInternal},
                    {DisplayName~Text : DisplayName},
                    {Price~Decimal : Price},
                    {Field~Text : Field},
                    {Result~Text : Result}
                }
                from @result;
         """,
        description="Find any outlier quotes for a given instrument using the interquartile range. "
        "This uses the 1.5 * IQR rule, that is, "
        "to find any quotes that fall between Q1 - 1.5 * IQR or above Q3 + 1.5 * IQR",
        parameters=[
            lumi.Parameter(
                name="quote_scope",
                value=deployment_name,
                set_as_default_value=True,
                tooltip="Scope to check quotes in",
                type=lumi.ParameterType.Text,
            )
        ],
    )
    quotes_update_view = lumi.ViewResource(
        id="quotes_update_view",
        provider=f"Views.fbnconfig.quote_writer_{deployment_name}",
        sql="""
            @@quote_scope = select #PARAMETERVALUE(quote_scope);
            @@instrument_id = select #PARAMETERVALUE(instrument_id);
            @@field_type = select #PARAMETERVALUE(field_type);
            @@price = select #PARAMETERVALUE(price);
            @@effective_at = select #PARAMETERVALUE(effective_at);

            @table_of_data =
                select
                    @@quote_scope as QuoteScope,
                    'Lusid' as Provider,
                    @@instrument_id as InstrumentId,
                    'ClientInternal' as InstrumentIdType,
                    'Price' as QuoteType,
                    @@field_type as Field,
                    @@price as Value,
                    'GBP' as Unit,
                    @@effective_at as QuoteEffectiveAt;

            select * from Lusid.Instrument.Quote.Writer where ToWrite = @table_of_data;
         """,
        description="I am a view",
        parameters=[
            lumi.Parameter(
                name="quote_scope",
                value=deployment_name,
                set_as_default_value=True,
                tooltip="Scope to check quotes in",
                type=lumi.ParameterType.Text,
            ),
            lumi.Parameter(
                name="instrument_id",
                value="",
                set_as_default_value=False,
                tooltip="Instrument id to load given quote for",
                type=lumi.ParameterType.Text,
            ),
            lumi.Parameter(
                name="field_type",
                value="",
                set_as_default_value=False,
                tooltip="Quote field type. Mid.",
                type=lumi.ParameterType.Text,
            ),
            lumi.Parameter(
                name="price",
                value=0,
                set_as_default_value=False,
                tooltip="Quote price",
                type=lumi.ParameterType.Decimal,
            ),
            lumi.Parameter(
                name="effective_at",
                value="2024-01-18",
                set_as_default_value=False,
                tooltip="Date to load the quote for",
                type=lumi.ParameterType.DateTime,
            ),
        ],
        use_dry_run=True,
    )

    import_from_excel_worker = wf.WorkerResource(
        id="import_from_excel_worker-example",
        scope=deployment_name,
        code="ImportFromExcelFile",
        display_name="Import From Excel",
        description="Imports quote data from specified Excel file in Drive.",
        worker_configuration=wf.LuminesceView(view=import_from_excel_view),
    )

    reasonable_value_worker = wf.WorkerResource(
        id="reasonable_value_worker-example",
        scope=deployment_name,
        code="ReasonableValueChecker",
        display_name="Reasonable Value Checker",
        description="Find any quotes with values not between 1 and 1000.",
        worker_configuration=wf.LuminesceView(view=reasonable_value_check_view),
    )

    iqr_outliers_worker = wf.WorkerResource(
        id="iqr_outliers_worker-example",
        scope=deployment_name,
        code="IQROutliers",
        display_name="IQR Outliers",
        description="Find any IQR outlier quotes.",
        worker_configuration=wf.LuminesceView(view=iqr_outlier_view),
    )

    quotes_update_worker = wf.WorkerResource(
        id="quotes_update_worker-example",
        scope=deployment_name,
        code="QuoteUpdate",
        display_name="Quote update",
        description="Update quote based on provided parameters",
        worker_configuration=wf.LuminesceView(view=quotes_update_view),
    )

    def external_trigger(name: str) -> wf.TriggerDefinition:
        return wf.TriggerDefinition(name=name, type="External")

    # Common triggers
    start_trigger = external_trigger(name="Start")
    resolved_trigger = external_trigger(name="Resolved")
    place_on_hold_trigger = external_trigger(name="PlaceOnHold")
    resume_trigger = external_trigger(name="Resume")
    ignore_trigger = external_trigger(name="Ignore")
    updated_trigger = external_trigger(name="Updated")
    no_exceptions_trigger = external_trigger(name="NoExceptions")
    exceptions_found_trigger = external_trigger(name="ExceptionsFound")

    # Common actions
    resolve_parent_action = wf.ActionDefinition(
        name="resolved-trigger-parent",
        action_details=wf.TriggerParentTaskAction(trigger=resolved_trigger),
    )

    # Common fields
    quote_scope_field: wf.TaskFieldDefinition = wf.TaskFieldDefinition(
        name="QuoteScope", type=wf.TaskFieldDefinitionType.STRING
    )

    # Common states
    pending_state = wf.TaskStateDefinition(name="Pending")
    in_dq_control_state = wf.TaskStateDefinition(name="InDQControl")
    exceptions_state = wf.TaskStateDefinition(name="Exceptions")
    complete_state = wf.TaskStateDefinition(name="Complete")
    resolved_state = wf.TaskStateDefinition(name="Resolved")
    ignored_state = wf.TaskStateDefinition(name="Ignored")
    in_progress_state = wf.TaskStateDefinition(name="InProgress")
    on_hold_state = wf.TaskStateDefinition(name="OnHold")
    # Common child task exception fields

    common_child_task_fields: dict[str | wf.TaskFieldDefinition, wf.FieldMapping] = {
        quote_scope_field: wf.FieldMapping(map_from=quote_scope_field),  # pyright: ignore
        "PriceDate": wf.FieldMapping(map_from="PriceDate"),
        "ClientInternal": wf.FieldMapping(map_from="ClientInternal"),
        "DisplayName": wf.FieldMapping(map_from="DisplayName"),
        "Price": wf.FieldMapping(map_from="Price"),
        "Result": wf.FieldMapping(map_from="Result"),
        "Field": wf.FieldMapping(map_from="Field"),
    }
    update_quote_action = wf.ActionDefinition(
        name="update-quote",
        action_details=wf.RunWorkerAction(
            worker=quotes_update_worker,
            worker_parameters={
                "quote_scope": wf.FieldMapping(map_from=quote_scope_field),
                "instrument_id": wf.FieldMapping(map_from="ClientInternal"),
                "field_type": wf.FieldMapping(map_from="Field"),
                "price": wf.FieldMapping(map_from="Price"),
                "effective_at": wf.FieldMapping(map_from="PriceDate"),
            },
            worker_status_triggers=wf.WorkerStatusTriggers(
                completed_with_results=updated_trigger, completed_no_results=place_on_hold_trigger
            ),
        ),
    )

    handle_exception_task_definition = wf.TaskDefinitionResource(
        id="handle_exception_task_definition",
        scope=deployment_name,
        code="HandleException",
        display_name="Handle Exception",
        description="Handle any data outliers that are raised.",
        states=[
            pending_state,
            in_progress_state,
            resolved_state,
            on_hold_state,
            ignored_state,
            complete_state,
        ],
        field_schema=[
            quote_scope_field,
            wf.TaskFieldDefinition(name="PriceDate", type=wf.TaskFieldDefinitionType.DATETIME),
            wf.TaskFieldDefinition(name="ClientInternal", type=wf.TaskFieldDefinitionType.STRING),
            wf.TaskFieldDefinition(name="DisplayName", type=wf.TaskFieldDefinitionType.STRING),
            wf.TaskFieldDefinition(name="Price", type=wf.TaskFieldDefinitionType.DECIMAL),
            wf.TaskFieldDefinition(name="Field", type=wf.TaskFieldDefinitionType.STRING),
            wf.TaskFieldDefinition(name="Result", type=wf.TaskFieldDefinitionType.STRING),
            wf.TaskFieldDefinition(name="Details", type=wf.TaskFieldDefinitionType.STRING),
        ],
        initial_state=wf.InitialState(
            name=pending_state,
            required_fields=[
                quote_scope_field,
                "PriceDate",
                "ClientInternal",
                "DisplayName",
                "Price",
                "Result",
            ],
        ),
        triggers=[
            start_trigger,
            resolved_trigger,
            place_on_hold_trigger,
            resume_trigger,
            ignore_trigger,
            updated_trigger,
        ],
        transitions=[
            wf.TaskTransitionDefinition(
                from_state=pending_state, to_state=in_progress_state, trigger=start_trigger
            ),
            wf.TaskTransitionDefinition(
                from_state=in_progress_state,
                to_state=resolved_state,
                trigger=resolved_trigger,
                guard="fields['Details'] neq ''",
                action=update_quote_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_progress_state, to_state=on_hold_state, trigger=place_on_hold_trigger
            ),
            wf.TaskTransitionDefinition(
                from_state=on_hold_state, to_state=in_progress_state, trigger=resume_trigger
            ),
            wf.TaskTransitionDefinition(
                from_state=in_progress_state,
                to_state=ignored_state,
                trigger=ignore_trigger,
                guard="fields['Details'] neq ''",
                action=resolve_parent_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=resolved_state,
                to_state=complete_state,
                trigger=updated_trigger,
                action=resolve_parent_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=resolved_state, to_state=on_hold_state, trigger=place_on_hold_trigger
            ),
        ],
        actions=[resolve_parent_action, update_quote_action],
    )

    start_reasonable_value_action = wf.ActionDefinition(
        name="start_reasonable_value_action",
        action_details=wf.RunWorkerAction(
            worker=reasonable_value_worker,
            worker_parameters={"quote_scope": wf.FieldMapping(map_from=quote_scope_field)},
            worker_status_triggers=wf.WorkerStatusTriggers(
                completed_with_results=exceptions_found_trigger,
                completed_no_results=no_exceptions_trigger,
            ),
            child_task_configurations=[
                wf.ResultantChildTaskConfiguration(
                    child_task_configuration=wf.ChildTaskConfiguration(
                        task_definition=handle_exception_task_definition,
                        initial_trigger=start_trigger,
                        child_task_fields=common_child_task_fields,
                    )
                )
            ],
        ),
    )

    reasonable_value_control_task = wf.TaskDefinitionResource(
        id="reasonable_value_control_task",
        scope=deployment_name,
        code="ReasonableValueDataControl",
        display_name="Reasonable Value Data Control",
        description="Conduct reasonable value data control on quote data. Raise any exceptions.",
        states=[pending_state, in_dq_control_state, exceptions_state, resolved_state, complete_state],
        field_schema=[quote_scope_field],
        initial_state=wf.InitialState(name=pending_state, required_fields=[quote_scope_field]),
        triggers=[start_trigger, no_exceptions_trigger, exceptions_found_trigger, resolved_trigger],
        actions=[start_reasonable_value_action, resolve_parent_action],
        transitions=[
            wf.TaskTransitionDefinition(
                from_state=pending_state,
                to_state=in_dq_control_state,
                trigger=start_trigger,
                action=start_reasonable_value_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_dq_control_state,
                to_state=complete_state,
                trigger=no_exceptions_trigger,
                action=resolve_parent_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_dq_control_state,
                to_state=exceptions_state,
                trigger=exceptions_found_trigger,
            ),
            wf.TaskTransitionDefinition(
                from_state=exceptions_state,
                to_state=resolved_state,
                guard=f"childTasks all (state eq '{complete_state.name}' "
                f"or state eq '{ignored_state.name}')",
                trigger=resolved_trigger,
                action=resolve_parent_action,
            ),
        ],
    )

    start_iqr_outlier_action = wf.ActionDefinition(
        name="start-iqr-outlier-action",
        action_details=wf.RunWorkerAction(
            worker=iqr_outliers_worker,
            worker_parameters={"quote_scope": wf.FieldMapping(map_from=quote_scope_field)},
            worker_status_triggers=wf.WorkerStatusTriggers(
                completed_with_results=exceptions_found_trigger,
                completed_no_results=no_exceptions_trigger,
            ),
            child_task_configurations=[
                wf.ResultantChildTaskConfiguration(
                    child_task_configuration=wf.ChildTaskConfiguration(
                        initial_trigger=start_trigger,
                        task_definition=handle_exception_task_definition,
                        child_task_fields=common_child_task_fields,
                    )
                )
            ],
        ),
    )

    iqr_outlier_control_task = wf.TaskDefinitionResource(
        id="iqr_outlier_control_task",
        scope=deployment_name,
        code="IQROutlierDataControl",
        display_name="IQR Outlier Data Control",
        description="Conduct IQR outlier data control on quote data. Raise any exceptions.",
        states=[pending_state, in_dq_control_state, exceptions_state, complete_state, resolved_state],
        field_schema=[quote_scope_field],
        initial_state=wf.InitialState(name=pending_state, required_fields=[quote_scope_field]),
        triggers=[start_trigger, no_exceptions_trigger, exceptions_found_trigger, resolved_trigger],
        actions=[start_iqr_outlier_action, resolve_parent_action],
        transitions=[
            wf.TaskTransitionDefinition(
                from_state=pending_state,
                to_state=in_dq_control_state,
                trigger=start_trigger,
                action=start_iqr_outlier_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_dq_control_state,
                to_state=complete_state,
                trigger=no_exceptions_trigger,
                action=resolve_parent_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_dq_control_state,
                to_state=exceptions_state,
                trigger=exceptions_found_trigger,
            ),
            wf.TaskTransitionDefinition(
                from_state=exceptions_state,
                to_state=resolved_state,
                guard=f"childTasks all (state eq '{complete_state.name}' "
                f"or state eq '{ignored_state.name}')",
                trigger=resolved_trigger,
                action=resolve_parent_action,
            ),
        ],
    )

    start_import_action = wf.ActionDefinition(
        name="start-import-worker",
        action_details=wf.RunWorkerAction(
            worker=import_from_excel_worker,
            worker_parameters={
                "quote_scope": wf.FieldMapping(map_from=quote_scope_field),
                "filename": wf.FieldMapping(map_from="filename"),
            },
            worker_status_triggers=wf.WorkerStatusTriggers(
                completed_with_results="Imported",
                completed_no_results="Failure",
                failed_to_complete="Failure",
                failed_to_start="Failure",
            ),
        ),
    )

    create_reasonable_value_task_action = wf.ActionDefinition(
        name="create-reasonable-value-task",
        action_details=wf.CreateChildTasksAction(
            child_task_configurations=[
                wf.ChildTaskConfiguration(
                    task_definition=reasonable_value_control_task,
                    initial_trigger=start_trigger,
                    child_task_fields={
                        quote_scope_field: wf.FieldMapping(map_from=quote_scope_field)  # pyright: ignore
                    },
                )
            ]
        ),
    )

    create_iqr_outlier_task_action = wf.ActionDefinition(
        name="create-IQR-outlier-task",
        action_details=wf.CreateChildTasksAction(
            child_task_configurations=[
                wf.ChildTaskConfiguration(
                    task_definition=iqr_outlier_control_task,
                    initial_trigger=start_trigger,
                    child_task_fields={
                        quote_scope_field: wf.FieldMapping(map_from=quote_scope_field)  # pyright: ignore
                    },
                )
            ]
        ),
    )

    main_task_definition = wf.TaskDefinitionResource(
        id="import_quotes_task_definition",
        scope=deployment_name,
        code="ImportQuotes",
        display_name="Import Quotes",
        description="Import and validate quote data from specified Excel file.",
        states=[
            pending_state,
            wf.TaskStateDefinition(name="ImportingQuotes"),
            wf.TaskStateDefinition(name="InReasonableValueDQControl"),
            wf.TaskStateDefinition(name="InIQROutlierDQControl"),
            complete_state,
            wf.TaskStateDefinition(name="Error"),
        ],
        field_schema=[
            quote_scope_field,
            wf.TaskFieldDefinition(name="filename", type=wf.TaskFieldDefinitionType.STRING),
        ],
        initial_state=wf.InitialState(
            name=pending_state, required_fields=[quote_scope_field, "filename"]
        ),
        triggers=[
            start_trigger,
            wf.TriggerDefinition(name="Failure", type="External"),
            wf.TriggerDefinition(name="Imported", type="External"),
            resolved_trigger,
        ],
        transitions=[
            wf.TaskTransitionDefinition(
                from_state=pending_state,
                to_state="ImportingQuotes",
                trigger=start_trigger,
                action=start_import_action,
            ),
            wf.TaskTransitionDefinition(
                from_state="ImportingQuotes",
                to_state="InReasonableValueDQControl",
                trigger="Imported",
                action=create_reasonable_value_task_action,
            ),
            wf.TaskTransitionDefinition(
                from_state="InReasonableValueDQControl",
                to_state="InIQROutlierDQControl",
                trigger=resolved_trigger,
                action=create_iqr_outlier_task_action,
            ),
            wf.TaskTransitionDefinition(
                from_state="ImportingQuotes", to_state="Error", trigger="Failure"
            ),
            wf.TaskTransitionDefinition(
                from_state="InIQROutlierDQControl",
                to_state=complete_state,
                trigger=resolved_trigger,
                guard=f"childTasks all (state eq '{resolved_state.name}')",
            ),
        ],
        actions=[
            start_import_action,
            create_reasonable_value_task_action,
            create_iqr_outlier_task_action,
        ],
    )

    return Deployment(deployment_name, [main_task_definition])



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import customentity

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeEntityTypeResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):
        respx_mock.post("/api/api/customentities/entitytypes").mock(
            return_value=httpx.Response(200, json={"entityType": "~typename"})
        )
        # given a desired definition with one field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        # then the state the typename returned by the create call
        assert state == {"entitytype": "~typename"}
        # and a create request was sent without the startValue
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/entitytypes"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "description": "",
                    "collectionType": "Single",
                    "required": False,
                }
            ],
        }

    def test_update_with_no_changes(self, respx_mock):
        # given an existing CE where the field has a description
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "legs",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "",
                            "collectionType": "Single",
                            "required": False,
                        }
                    ],
                },
            )
        )
        # and a desired which is the same but the field description is none
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and a read was made but no PUT

    def test_update_with_collection_single(self, respx_mock):
        # given an existing CE with a collection type of single
        # note, api will not return single here even if the user has
        # set it
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "required": False,
                            # no collectionType member from GET
                        }
                    ],
                },
            )
        )
        # and a desired where the user has explicitly asked for single
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="arms",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    collection_type=customentity.CollectionType.SINGLE,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state None because there is no change
        assert state is None

    def test_update_with_changed_field(self, respx_mock):
        # given an existing CE with an arms field
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        }
                    ],
                },
            )
        )
        respx_mock.put("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(200, json={"entityType": "~whatevah"})
        )
        # and a desired with a legs field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is the same
        assert state == {"entitytype": "~whatevah"}
        # and a put request was sent with legs
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/customentities/entitytypes/~whatevah"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                }
            ],
        }

    def test_update_with_removed_field(self, respx_mock):
        # given an existing CE with arms and legs
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "legs",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        },
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        },
                    ],
                },
            )
        )
        respx_mock.put("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(200, json={"entityType": "~whatevah"})
        )
        # and a desired with a legs field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is the same
        assert state == {"entitytype": "~whatevah"}
        # and a put request was sent with legs
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/customentities/entitytypes/~whatevah"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                }
            ],
        }

    def test_update_with_additional_field(self, respx_mock):
        # given an existing CE with an arms field
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        }
                    ],
                },
            )
        )
        respx_mock.put("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(200, json={"entityType": "~whatevah"})
        )
        # and a desired with an arms and a legs field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                ),
                customentity.FieldDefinition(
                    name="arms",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                ),
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is the same
        assert state == {"entitytype": "~whatevah"}
        # and a put request was sent with the new fields
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/customentities/entitytypes/~whatevah"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                },
                {
                    "name": "arms",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                },
            ],
        }

    def test_delete_throws(self):
        # given a resource that exists in the remnte
        old_state = SimpleNamespace(entitytype="~whatever")
        # when we delete it throws brcause uou cant delete a CE
        with pytest.raises(RuntimeError):
            customentity.EntityTypeResource.delete(self.client, old_state)

    def test_deps(self):
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[],
        )
        # it's deps are empty
        assert sut.deps() == []



import os
from types import SimpleNamespace

from pytest import fixture

import fbnconfig
from fbnconfig import reference_list as rl
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture()
def base_resources(setup_deployment):
    deployment_name = setup_deployment.name
    # Create a simple string list reference list
    string_reflist = rl.ReferenceListResource(
        id="string_reflist",
        scope=deployment_name,
        code="string-list-1",
        name="Test String Reference List",
        description="A test reference list with string values",
        tags=["test", "integration"],
        reference_list=rl.StringList(values=["value1", "value2", "value3"])
    )
    # Create a decimal list reference list
    decimal_reflist = rl.ReferenceListResource(
        id="decimal_reflist",
        scope=deployment_name,
        code="decimal-list-1",
        name="Test Decimal Reference List",
        description="A test reference list with decimal values",
        reference_list=rl.DecimalList(values=[1.5, 2.75, 3.0])
    )
    # Create an address key list reference list
    address_reflist = rl.ReferenceListResource(
        id="address_reflist",
        scope=deployment_name,
        code="address-key-list-1",
        name="Test Address Key Reference List",
        reference_list=rl.AddressKeyList(values=["Portfolio/Name", "Portfolio/Currency"])
    )
    return [string_reflist, decimal_reflist, address_reflist]


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("reference_list")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Clean up reference lists after tests
    try:
        client.delete(f"/api/api/referencelists/{deployment_name}/string-list-1")
        client.delete(f"/api/api/referencelists/{deployment_name}/decimal-list-1")
        client.delete(f"/api/api/referencelists/{deployment_name}/address-key-list-1")
    except Exception:
        pass  # Ignore cleanup errors


def test_create(setup_deployment, base_resources):
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # Verify string reference list was created
    string_response = client.get(f"/api/api/referencelists/{setup_deployment.name}/string-list-1").json()
    assert string_response["id"]["scope"] == setup_deployment.name
    assert string_response["id"]["code"] == "string-list-1"
    assert string_response["name"] == "Test String Reference List"
    assert string_response["description"] == "A test reference list with string values"
    assert string_response["tags"] == ["test", "integration"]
    assert string_response["referenceList"]["values"] == ["value1", "value2", "value3"]
    assert string_response["referenceList"]["referenceListType"] == "StringList"
    # Verify decimal reference list was created
    decimal_response = client.get(
        f"/api/api/referencelists/{setup_deployment.name}/decimal-list-1"
    ).json()
    assert decimal_response["id"]["scope"] == setup_deployment.name
    assert decimal_response["id"]["code"] == "decimal-list-1"
    assert decimal_response["name"] == "Test Decimal Reference List"
    assert decimal_response["referenceList"]["values"] == [1.5, 2.75, 3.0]
    assert decimal_response["referenceList"]["referenceListType"] == "DecimalList"
    # Verify address key reference list was created
    address_response = client.get(
        f"/api/api/referencelists/{setup_deployment.name}/address-key-list-1"
    ).json()
    assert address_response["id"]["scope"] == setup_deployment.name
    assert address_response["id"]["code"] == "address-key-list-1"
    assert address_response["name"] == "Test Address Key Reference List"
    assert address_response["referenceList"]["values"] == ["Portfolio/Name", "Portfolio/Currency"]
    assert address_response["referenceList"]["referenceListType"] == "AddressKeyList"


def test_nochange(setup_deployment, base_resources):
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # when we apply it again
    update = fbnconfig.deploy(deployment, lusid_env, token)
    # then there are no changes
    ref_list_changes = [a.change for a in update if a.type == "ReferenceListResource"]
    assert ref_list_changes == ["nochange", "nochange", "nochange"]


def test_teardown(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # when we remove all the resources
    empty = fbnconfig.Deployment(deployment_name, [])
    update = fbnconfig.deploy(empty, lusid_env, token)
    # then there are no changes
    ref_list_changes = [a.change for a in update if a.type == "ReferenceListResource"]
    assert ref_list_changes == ["remove", "remove", "remove"]


def test_update(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # Given we have deployed the base case
    initial = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deploy(initial, lusid_env, token)
    # when we update a resource
    updated_resources = [
        # Update the string list with different values
        rl.ReferenceListResource(
            id="string_reflist",
            scope=deployment_name,
            code="string-list-1",
            name="Updated String Reference List",  # Changed name
            description="An updated test reference list with string values",  # Changed description
            tags=["test", "integration", "updated"],  # Added tag
            reference_list=rl.StringList(
                values=["newvalue1", "newvalue2"]
            )  # Changed values
        ),
        # Keep decimal list the same
        rl.ReferenceListResource(
            id="decimal_reflist",
            scope=deployment_name,
            code="decimal-list-1",
            name="Test Decimal Reference List",
            description="A test reference list with decimal values",
            reference_list=rl.DecimalList(values=[1.5, 2.75, 3.0])
        ),
        # Keep address key list the same
        rl.ReferenceListResource(
            id="address_reflist",
            scope=deployment_name,
            code="address-key-list-1",
            name="Test Address Key Reference List",
            reference_list=rl.AddressKeyList(values=["Portfolio/Name", "Portfolio/Currency"])
        ),
    ]
    # and deploy it
    updated_deployment = fbnconfig.Deployment(deployment_name, updated_resources)  # type: ignore
    update = fbnconfig.deploy(updated_deployment, lusid_env, token)
    # then we expect only the modified resource to change
    ref_list_changes = [a.change for a in update if a.type == "ReferenceListResource"]
    assert ref_list_changes == ["update", "nochange", "nochange"]
    # and it has the new values
    updated_response = client.get(
        f"/api/api/referencelists/{deployment_name}/string-list-1"
    ).json()
    assert updated_response["name"] == "Updated String Reference List"
    assert updated_response["description"] == "An updated test reference list with string values"
    assert updated_response["tags"] == ["test", "integration", "updated"]
    assert updated_response["referenceList"]["values"] == ["newvalue1", "newvalue2"]



import pathlib

from fbnconfig import Deployment, drive, lumi

"""
An example configuration for defining Luminesce views.
The script configures the following entities:
- Folder
- File
- View

For more information on Folder and File resources, please refer to the `drive.py`
example.

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01767/
"""


def configure(env):
    lumi.ViewResource.Registration.wait_time = 30
    xlsx = "../data/example-spreadsheet.xlsx"
    base_folder = drive.FolderResource(id="base_folder", name="fbnconfig-folder", parent=drive.root)
    deployment_folder = drive.FolderResource(
        id="sub_folder", name="deployment-folder", parent=base_folder
    )
    content_path = pathlib.Path(__file__).parent.resolve() / pathlib.Path(xlsx)
    spreadsheet = drive.FileResource(
        id="xslx", folder=deployment_folder, name="example-spreadsheet.xlsx", content_path=content_path
    )
    xlsx_sql = f"""\
        -- Pull data from the spreadsheet
        @client_register_master = use Drive.Excel
        --file={spreadsheet.path()}
        enduse;

        select * from @client_register_master;
    """
    xls_view = lumi.ViewResource(
        id="xls-view",
        provider="Views.fbnconfig.ExampleProvider",
        description="Test view for fbnconfig reading a spreadsheet",
        sql=xlsx_sql,
        dependencies=[spreadsheet],
        use_dry_run=True,
    )
    vr_provider = "Views.fbnconfig.ExampleProvider_vr"
    vr = lumi.ViewResource(
        id="lumi-example-view",
        provider=vr_provider,
        description="My resource test view",
        documentation_link="http://example.com/query4",
        variable_shape=False,
        use_dry_run=True,
        allow_execute_indirectly=False,
        distinct=True,
        sql="""
            select 2+#PARAMETERVALUE(p1)
            as   twelve

        """,
        parameters=[
            lumi.Parameter(
                name="p1",
                value=10,
                set_as_default_value=True,
                tooltip="a number, defaults to 10",
                type=lumi.ParameterType.Int,
            ),
            lumi.Parameter(
                name="p2",
                value="a default",
                set_as_default_value=True,
                tooltip="a string",
                type=lumi.ParameterType.Text,
            ),
        ],
        dependencies=[],
    )
    v2_provider = "Views.fbnconfig.ExampleProvider_v2"
    v2 = lumi.ViewResource(
        id="lumi-dependent-view",
        provider=v2_provider,
        description="My resource test view",
        documentation_link="http://example.com/query4",
        variable_shape=False,
        use_dry_run=True,
        allow_execute_indirectly=False,
        distinct=True,
        sql=f"select twelve from {vr_provider}",
        parameters=[
            lumi.Parameter(
                name="num",
                value=10,
                set_as_default_value=False,
                tooltip="a number",
                type=lumi.ParameterType.Int,
            )
        ],
        dependencies=[vr],
    )
    seven = lumi.Variable(name="seven", type=lumi.VariableType.Table, sql="select 3 + 4")
    scalar_var = lumi.ViewResource(
        id="lumi-scalar-var-view",
        provider="Views.fbnconfig.ExampleProvider_scalar_var",
        description="View which takes a scalar variable",
        documentation_link="http://example.com/vars_query",
        use_dry_run=False,
        distinct=True,
        sql="select * from #PARAMETERVALUE(num)",
        variables=[seven],
        parameters=[
            lumi.Parameter(
                name="num",
                value=seven,
                is_mandatory=False,
                tooltip="this should be seven",
                type=lumi.ParameterType.Table,
            )
        ],
    )
    return Deployment("luminesce_example", [vr, v2, xls_view, scalar_var])



import json
from types import SimpleNamespace
from typing import List, NamedTuple, Union

import httpx

from fbnconfig import http_client, log
from fbnconfig.resource_abc import Ref, Resource, get_resource_class, get_resource_type  # noqa: F401

Deployment = NamedTuple("Deployment", [("id", str), ("resources", List[Union[Resource, Ref]])])
deployments = []

Action = NamedTuple("Action", [("id", str), ("type", str), ("change", str)])


def deployment(id, resources):
    d = Deployment(id, resources)
    deployments.append(d)


def recurse_resources(resources):
    seen = dict()
    ordered = []

    def helper(items):
        for res in items:
            deps = res.deps()
            if len(deps) > 0:
                helper(deps)
            if seen.get(res.id, None) is None:
                seen[res.id] = id(res)
                ordered.append(res)
            elif id(res) != seen[res.id]:
                raise RuntimeError(f"Resource ID: {res.id} has been used on more than one resource")

    helper(resources)
    return ordered


def sort_deployed(entries):
    seen = set()
    ordered = []

    def helper(items):
        for e in items:
            dep_ids = e.dependencies
            if len(dep_ids) > 0:
                deps = [e for e in entries if e.resource_id in dep_ids]
                helper(deps)
            if e.resource_id not in seen:
                seen.add(e.resource_id)
                ordered.append(e)

    helper(entries)
    return ordered


def run(client: httpx.Client, deployment) -> List[Action]:
    actions = []
    resources = deployment.resources
    all_resources = recurse_resources(resources)
    deploy_state = log.list_resources_for_deployment(client, deployment.id)
    for resource in all_resources:
        if isinstance(resource, Ref):
            resource.attach(client)
            actions.append(Action(id=resource.id, type=get_resource_type(resource), change="attach"))
            print("attach", get_resource_type(resource))
            continue
        match = [e for e in deploy_state if e.resource_id == resource.id]
        if len(match) == 0:
            print("deploy: create", get_resource_type(resource), resource.id)
            state = resource.create(client)
            log.record(client, deployment.id, resource, state)
            actions.append(Action(id=resource.id, type=get_resource_type(resource), change="create"))
        else:
            print("deploy: update", get_resource_type(resource), resource.id)
            state = resource.update(client, match[0].state)
            if state is not None:
                log.record(client, deployment.id, resource, state)
                actions.append(Action(id=resource.id, type=get_resource_type(resource), change="update"))
            else:
                print("deploy: nochange", get_resource_type(resource), resource.id)
                actions.append(
                    Action(id=resource.id, type=get_resource_type(resource), change="nochange")
                )
    for deployed in reversed(sort_deployed(deploy_state)):
        matching_resource = [e for e in all_resources if e.id == deployed.resource_id]
        if len(matching_resource) == 0:
            klas = get_resource_class(deployed.resource_type)
            print("deploy: remove", deployed.resource_type, deployed.resource_id)
            klas.delete(client, deployed.state)
            log.remove(client, deployment.id, deployed.resource_id)
            actions.append(Action(id=deployed.resource_id, type=deployed.resource_type, change="remove"))
    return actions


# called as a library
def deploy(deployment: Deployment, lusid_url: str, access_token: str) -> List[Action]:
    client = http_client.create_client(lusid_url, access_token)
    return run(client, deployment)


def load_vars(vars_file):
    if vars_file:
        host_vars = SimpleNamespace(**json.load(vars_file))
    else:
        host_vars = SimpleNamespace()
    return host_vars
